# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from pinball_ext.job_templates import CommandConditionTemplate
from pinball_ext.job_templates import CommandJobTemplate
from pinball_ext.job_templates import JobTemplate


class JobTemplateTestCase(unittest.TestCase):
    def test_job_template(self):
        job_template = JobTemplate(
            'MyJobClassName',
            'qubole',
            {
                'USER_APP_JAR': '/my/app/jar.jar',
            })
        pb_job = job_template.get_pinball_job(
            [],
            [],
            {
                'end_date': '2015-02-25',
                'job_repo_dir': '/mnt/job_repo/',
                'job_import_dirs_config': 'pinball_ext.example.JOB_IMPORT_DIRS',
            })
        self.assertEqual(
            pb_job.command,
            'cd /mnt/job_repo/ && '
            'python -m pinball_ext.job.job_runner '
            '--job_import_dirs_config=pinball_ext.example.JOB_IMPORT_DIRS '
            '--job_class_name=MyJobClassName '
            '--job_params="end_date=2015-02-25" '
            '--executor=qubole '
            '--executor_config="USER_APP_JAR=/my/app/jar.jar" ')


class CommandJobTemplatesTestCase(unittest.TestCase):
    def test_command_job_template(self):
        """Test the config produced by a command job template."""
        job_template = CommandJobTemplate('some_job', 'some_command')
        pb_job = job_template.get_pinball_job([], [], params={
           'end_date': '2015-02-25',
        })
        self.assertEqual('some_command', pb_job.command)


class ConditionTemplatesTestCase(unittest.TestCase):
    def test_command_condition_template(self):
        """Test the config produced by a command condition template."""
        condition = CommandConditionTemplate('some_condition')
        condition_params = {'command': 'some_command'}
        condition_job = condition.get_pinball_condition(None, condition_params)
        self.assertEqual('some_command', condition_job.command)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for oauth2"""
import json
import unittest

from pinball.authentication.oauth2 import Crypter, CryptoException
from pinball.config.pinball_config import PinballConfig


__author__ = 'Tongbo Huang, Devin Lundberg'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Tongbo Huang', 'Devin Lundberg']
__license__ = 'Apache'
__version__ = '2.0'


class CrypterTestCase(unittest.TestCase):
    """Test cases for Crypter"""
    plaintext = 'secret'
    arc2_ciphertext = '\x1a\x9c^\xbc\xe5\xef\xf9\x97'

    def setUp(self):
        self.crypter = Crypter()

    def test_encryption(self):
        """Encrypt and decrypt a test string."""

        encrypted_string = self.crypter.encrypt(self.plaintext)
        decrypted_string = self.crypter.decrypt(encrypted_string)

        self.assertNotEqual(encrypted_string, self.plaintext)
        self.assertEqual(decrypted_string, self.plaintext)

    def test_nondeterministic_encryption(self):
        """Makes sure encryption is nondeterministic"""
        encrypted_string = self.crypter.encrypt(self.plaintext)
        encrypted_string2 = self.crypter.encrypt(self.plaintext)
        self.assertNotEqual(encrypted_string, encrypted_string2)

    def test_legacy_encryption(self):
        """Decrypt arc2 ciphertext (for legacy support)"""
        decrypted_string = self.crypter.decrypt(self.arc2_ciphertext)

        self.assertEqual(decrypted_string, self.plaintext)

    def test_serialize_deserialize(self):
        ciphertext = 'ciphertext'
        auth = 'auth'
        test = 'test'
        out = self.crypter._serialize(ciphertext, auth, test=test)
        v, ret_ciphertext, ret_auth, params = self.crypter._deserialize(out)

        self.assertEqual(v, PinballConfig.CRYPTO_VERSION)
        self.assertEqual(ciphertext, ret_ciphertext)
        self.assertEqual(auth, ret_auth)
        self.assertEqual(1, len(params))
        self.assertTrue('test' in params)
        self.assertEqual(params['test'], test)

    def test_deserialize_encoding_error(self):
        unencoded_data = json.dumps({
            'ciphertext': '',
            'auth': '',
            'params': {}
        })
        with self.assertRaises(CryptoException):
            self.crypter._deserialize(unencoded_data)

    def test_deserialize_not_json_error(self):
        not_json_data = 'This is not json'.encode('base64')
        with self.assertRaises(CryptoException):
            self.crypter._deserialize(not_json_data)

    def test_deserialize_wrong_json_error(self):
        serialized_without_version = json.dumps({
            'ciphertext': '',
            'auth': '',
            'params': {}
        }).encode('base64')
        with self.assertRaises(CryptoException):
            self.crypter._deserialize(serialized_without_version)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for schedules."""
import mock
import unittest

from pinball.master.thrift_lib.ttypes import Token
from pinball.parser.config_parser import PARSER_CALLER_KEY
from pinball.scheduler.overrun_policy import OverrunPolicy
from pinball.scheduler.schedule import Schedule
from pinball.scheduler.schedule import WorkflowSchedule
from pinball.ui.data import Status
from pinball.ui.data import WorkflowData
from pinball.ui.data import WorkflowInstanceData
from pinball.workflow.emailer import Emailer
from pinball.workflow.name import Name
from pinball.workflow.signaller import Signal
from tests.pinball.persistence.ephemeral_store import EphemeralStore


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class ScheduleTestCase(unittest.TestCase):
    class FakeSchedule(Schedule):
        def run(self, emailer, store):
            return None

        def is_running(self):
            return False

        def is_failed(self):
            return False

        def abort_running(self, client, store):
            return

    @mock.patch('time.time')
    def test_advance_next_run_time(self, time_mock):
        time_mock.return_value = 100.  # the value must be a float
        schedule = ScheduleTestCase.FakeSchedule(next_run_time=95,
                                                 recurrence_seconds=10)
        schedule.advance_next_run_time()
        self.assertTrue(105, schedule.next_run_time)

    def test_corresponds_to(self):
        some_schedule = ScheduleTestCase.FakeSchedule(next_run_time=10,
                                                      recurrence_seconds=10)
        corresponding_schedule = ScheduleTestCase.FakeSchedule(
            next_run_time=20, recurrence_seconds=10)
        self.assertTrue(some_schedule.corresponds_to(corresponding_schedule))
        non_corresponding_schedule = ScheduleTestCase.FakeSchedule(
            next_run_time=25, recurrence_seconds=10)
        self.assertFalse(some_schedule.corresponds_to(
                         non_corresponding_schedule))
        other_non_corresponding_schedule = ScheduleTestCase.FakeSchedule(
            next_run_time=10, recurrence_seconds=10,
            overrun_policy=OverrunPolicy.DELAY)
        self.assertFalse(some_schedule.corresponds_to(
                         other_non_corresponding_schedule))


class WorkflowScheduleTestCase(unittest.TestCase):
    def test_corresponds_to(self):
        some_schedule = WorkflowSchedule(next_run_time=20,
                                         recurrence_seconds=10,
                                         workflow='some_workflow')
        corresponding_schedule = WorkflowSchedule(next_run_time=10,
                                                  recurrence_seconds=10,
                                                  workflow='some_workflow')
        self.assertTrue(some_schedule.corresponds_to(corresponding_schedule))
        non_corresponding_schedule = WorkflowSchedule(
            next_run_time=20, recurrence_seconds=10,
            workflow='some_other_workflow')
        self.assertFalse(some_schedule.corresponds_to(
                         non_corresponding_schedule))

    @mock.patch('pinball.scheduler.schedule.load_path')
    def test_run(self, load_path_mock):
        config_parser = mock.Mock()

        def load_path(params):
            self.assertEqual([PARSER_CALLER_KEY], params.keys())
            return config_parser
        load_path_mock.return_value = load_path
        name = Name(workflow='some_workflow',
                    instance='123',
                    job_state=Name.WAITING_STATE,
                    job='some_job')
        config_parser.get_workflow_tokens.return_value = [
            Token(name=name.get_job_token_name())]

        schedule = WorkflowSchedule(workflow='some_workflow')
        store = EphemeralStore()
        emailer = Emailer('some_host', '8080')
        request = schedule.run(emailer, store)

        self.assertEqual(1, len(request.updates))

    @mock.patch('pinball.scheduler.schedule.DataBuilder')
    def test_is_running(self, data_builder_mock):
        store = mock.Mock()
        data_builder = mock.Mock()
        data_builder_mock.return_value = data_builder

        schedule = WorkflowSchedule(workflow='some_workflow')

        data_builder.get_workflow.return_value = None
        self.assertFalse(schedule.is_running(store))

        workflow_data = WorkflowData('some_workflow', status=Status.RUNNING)
        data_builder.get_workflow.return_value = workflow_data
        self.assertTrue(schedule.is_running(store))

    @mock.patch('pinball.scheduler.schedule.DataBuilder')
    def test_is_failed(self, data_builder_mock):
        store = mock.Mock()
        data_builder = mock.Mock()
        data_builder_mock.return_value = data_builder

        schedule = WorkflowSchedule(workflow='some_workflow')
        data_builder.get_workflow.return_value = None
        self.assertFalse(schedule.is_failed(store))

        workflow_data = WorkflowData('some_workflow', status=Status.FAILURE)
        data_builder.get_workflow.return_value = workflow_data
        self.assertTrue(schedule.is_failed(store))

    @mock.patch('pinball.scheduler.schedule.Signaller')
    @mock.patch('pinball.scheduler.schedule.DataBuilder')
    def test_abort_running(self, data_builder_mock, signaller_mock):
        client = mock.Mock()
        store = mock.Mock()

        data_builder = mock.Mock()
        data_builder_mock.return_value = data_builder

        schedule = WorkflowSchedule(workflow='some_workflow')
        failed_instance = WorkflowInstanceData('some_workflow',
                                               '123',
                                               status=Status.FAILURE)
        running_instance = WorkflowInstanceData('some_workflow',
                                                '12345',
                                                status=Status.RUNNING)

        data_builder.get_instances.return_value = [failed_instance,
                                                   running_instance]

        signaller = mock.Mock()
        signaller_mock.return_value = signaller

        schedule.abort_running(client, store)

        signaller_mock.assert_called_once_with(client,
                                               workflow='some_workflow',
                                               instance='12345')
        signaller.set_action.assert_called_once_with(Signal.ABORT)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parser that reads configs from a repository."""
import collections
import pickle

from pinball.config.pinball_config import PinballConfig
from pinball.master.thrift_lib.ttypes import Token
from pinball.parser.config_parser import ConfigParser
from pinball.parser.utils import recurrence_str_to_sec
from pinball.parser.utils import schedule_to_timestamp
from pinball.repository.github_repository import GithubRepository
from pinball.scheduler.overrun_policy import OverrunPolicy
from pinball.scheduler.schedule import WorkflowSchedule
from pinball.workflow.event import Event
from pinball.workflow.name import Name
from pinball.workflow.utils import get_unique_workflow_instance
from pinball.workflow.utils import load_path


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class RepositoryConfigParser(ConfigParser):
    def __init__(self, params=None):
        self._repository = GithubRepository()

    @staticmethod
    def _job_config_to_job(job_config, outputs):
        """Create a job from a job config.

        Args:
            job_config: The job config.
            outputs: The list of job output names.
        Returns:
            Pinball job constructed from the config.
        """
        # load template
        job_template = load_path(job_config.template)(
            job_config.job,
            max_attempts=job_config.max_attempts,
            emails=job_config.emails,
            priority=job_config.priority)
        return job_template.get_pinball_job(job_config.parents,
                                            outputs,
                                            job_config.template_params)

    @staticmethod
    def _condition_config_to_condition(condition_config, outputs):
        """Create a condition from a condition config.

        Args:
            condition_config: The condition config.
            outputs: The list of condition output names.
        Returns:
            Pinball job constructed from the config.
        """
        # load template
        condition_template = load_path(condition_config.template)(
            condition_config.job,
            max_attempts=condition_config.max_attempts,
            retry_delay_sec=condition_config.retry_delay_sec,
            emails=condition_config.emails,
            priority=condition_config.priority)
        return condition_template.get_pinball_condition(
            outputs, params=condition_config.template_params)

    @staticmethod
    def _job_config_to_token(workflow, instance, job_config, job_outputs):
        """Create a job token from a job config.

        Args:
            workflow: The workflow name.
            instance: The workflow instance.
            job_config: The job config to create token from.
            job_outputs: The names of the job outputs.
        Returns:
            Job token constructed from the job config.
        """
        if job_config.is_condition:
            job = RepositoryConfigParser._condition_config_to_condition(
                job_config, job_outputs)
        else:
            job = RepositoryConfigParser._job_config_to_job(job_config,
                                                            job_outputs)
        name = Name(workflow=workflow, instance=instance,
                    job_state=Name.WAITING_STATE, job=job_config.job)
        job_token = Token(name=name.get_job_token_name(),
                          data=pickle.dumps(job))
        return job_token

    def get_schedule_token(self, workflow):
        schedule_config = self._repository.get_schedule(workflow)
        timestamp = schedule_to_timestamp(schedule_config.time,
                                          schedule_config.start_date)
        recurrence = recurrence_str_to_sec(schedule_config.recurrence)
        overrun_policy = OverrunPolicy.from_string(
            schedule_config.overrun_policy)
        schedule = WorkflowSchedule(
            next_run_time=timestamp,
            recurrence_seconds=recurrence,
            overrun_policy=overrun_policy,
            workflow=schedule_config.workflow,
            emails=schedule_config.emails,
            #TODO(mao): to make it flexible that allow users specify through UI
            max_running_instances=PinballConfig.DEFAULT_MAX_WORKFLOW_RUNNING_INSTANCES)
        schedule.advance_next_run_time()
        timestamp = schedule.next_run_time
        token_name = (
            Name(workflow=schedule_config.workflow
                 ).get_workflow_schedule_token_name())
        return Token(name=token_name, owner='parser',
                     expirationTime=timestamp,
                     data=pickle.dumps(schedule))

    def get_workflow_tokens(self, workflow):
        # TODO(pawel): add workflow connectivity check.
        job_configs = {}
        top_level_job_names = []
        job_names = self._repository.get_job_names(workflow)
        for job_name in job_names:
            job_config = self._repository.get_job(workflow, job_name)
            job_configs[job_name] = job_config
            if not job_config.parents:
                top_level_job_names.append(job_name)
                job_config.parents = [Name.WORKFLOW_START_INPUT]

        job_outputs = collections.defaultdict(list)
        for job_config in job_configs.values():
            for parent_job_name in job_config.parents:
                job_outputs[parent_job_name].append(job_config.job)

        result = []
        instance = get_unique_workflow_instance()

        # Convert job configs to job tokens.
        for job_config in job_configs.values():
            token = RepositoryConfigParser._job_config_to_token(
                workflow,
                instance,
                job_config,
                job_outputs[job_config.job])
            result.append(token)

        # Create triggering events for top-level jobs.
        for job_name in top_level_job_names:
            event = Event(creator='repository_config_parser')
            event_name = Name(workflow=workflow,
                              instance=instance,
                              job=job_name,
                              input_name=Name.WORKFLOW_START_INPUT,
                              event='workflow_start_event')
            result.append(Token(name=event_name.get_event_token_name(),
                                data=pickle.dumps(event)))

        return result

    def get_workflow_names(self):
        return self._repository.get_workflow_names()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An executor to run Hadoop job in Amazon Elastic MapReduce cluster.

For more information about EMR, check: http://aws.amazon.com/elasticmapreduce/
"""

import datetime
import os
import re
import subprocess

from pinball_ext.common import hadoop_utils
from pinball_ext.common import utils
from pinball_ext.common.decorators import retry
from pinball_ext.executor.cluster_executor import ClusterExecutor
from pinball_ext.executor.common import Platform


__author__ = 'Changshu Liu, Zach Drach'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class EMRExecutor(ClusterExecutor):
    """Implements the executor on the EMR platform.

    The config.USER_JAR_DIRS and config.USER_APP_JAR are local paths relative to
    self.config.HADOOP_HOST_HOME/self.config.USER on machine:
    self.config.HADOOP_HOST_NAME (usually, it's the EMR master node).

    So you need to upload everything to that folder before executing any Hadoop
    job using this executor. Usually, those jars are built and uploaded to
    corresponding folder automatically as part of build system.

    TODO(csliu): make run_hadoop_streaming_job() work.
    """

    class Config(ClusterExecutor.Config):
        # Directory that hive query files are put into on hadoop master.
        # The hive query itself as well as its output will be in a custom
        # directory under this folder.
        HIVE_QUERIES_DIR = '/mnt/tmp'

        HIVE_FAILED_PREFIX = 'FAILED:'
        HIVE_HADOOP_FAILURE = ['java.lang.RuntimeException',  # Java error
                               'Processes ended with exit',  # Java error
                               'Program will exit.',  # Java error
                               'Exception in thread',  # Java java error
                               'does not exist']  # Hive add jar error

        SCHEDULER_PARAM = 'mapred.job.queue.name'

        PLATFORM = Platform.EMR

    def run_hive_query(self, query_str, upload_archive=False):
        full_query_string = \
            self._generate_hive_query_header(upload_archive=upload_archive)
        full_query_string += self._get_scheduler_hive_setting()
        full_query_string += query_str

        self.log.info('Running query:\n %s' % full_query_string)

        query_dir = self._mk_tmp_dir()
        hadoop_utils.put_string_to_hadoop(
            self.hadoop_host_config,
            full_query_string,
            '%s/query.ql' % query_dir)

        hadoop_utils.run_and_check_command_in_hadoop(
            self.hadoop_host_config,
            'set -o pipefail; /home/hadoop/hive/bin/hive -f %s/query.ql 2>&1 '
            '> %s/out.csv | tee %s/out.err'
            % (query_dir, query_dir, query_dir),
            log_line_processor=self._hive_query_log_line_processor)

        q_stdout = self._get_raw_query_result('%s/out.csv' % query_dir)
        q_stderr = self._get_raw_query_result('%s/out.err' % query_dir)

        self._check_for_hive_failure_message(q_stderr)

        self.log.info("Output has %d rows. First 10 rows:\n\t%s"
                      % (len(q_stdout),
                         '\n\t'.join([str(o) for o in q_stdout[:9]])))

        return q_stdout, q_stderr, self.job_ids

    def _check_for_hive_failure_message(self, q_stderr):
        """Catch when hive fails without returning a non-zero error code.

        q_stderr: a list of lines from q_stderr.
        """
        failed = any(' '.join(line).startswith(self.config.HIVE_FAILED_PREFIX)
                     for line in q_stderr)
        failed = failed or any(
            any(f in line for f in self.config.HIVE_HADOOP_FAILURE)
            in line for line in q_stderr)
        if failed:
            # hive didn't actually returns a non-zero error code.
            raise subprocess.CalledProcessError(
                1,
                '\n'.join('\t'.join(l) for l in q_stderr))

    def _hive_query_log_line_processor(self, line):
        """ A callback function that gets executed for every line of stderr
        coming from the running job.

        Returns a dict of pinball metadata.
        """
        job_regex = \
            r'Starting Job = (?P<job_id>\w+), Tracking URL = (?P<job_url>.+)'
        m = re.search(job_regex, line)
        if m:
            job_id = m.group('job_id')
            job_url = m.group('job_url')
            if job_id and job_url:
                self.job_ids.append(job_id)
                return {'job_id': job_id,
                        'job_url': job_url,
                        'kill_id': '%s/%s' % (self.config.PLATFORM, job_id)}
        return {}

    @retry(subprocess.CalledProcessError)
    def _mk_tmp_dir(self):
        query_timestamp =\
            datetime.datetime.utcnow().strftime("%Y-%m-%d-%H-%M-%S")
        query_dir = os.path.join(
            self.config.HIVE_QUERIES_DIR,
            self.config.USER,
            '%s_%s_%s' % (query_timestamp,
                          utils.get_random_string(),
                          self.config.NAME))
        hadoop_utils.run_and_check_command_in_hadoop(
            self.hadoop_host_config,
            command='mkdir -p %s' % query_dir)
        return query_dir

    @retry(subprocess.CalledProcessError)
    def _get_raw_query_result(self, output_file):
        """Return the raw query output from a query output file.

        Args:
            output_file: Location on the Hadoop local master of query output.

        Returns:
            A list of parsed rows. Each row is a list of strings.
        """
        rows = []
        with hadoop_utils.run_command_in_hadoop(
                self.hadoop_host_config,
                command='cat %s' % output_file) as hive_out:
            for line in hive_out:
                if line.strip():
                    rows.append(line.strip().split('\t'))
        return rows

    def kill_job(self, job_id):
        """Kills a EMR job with the given job_id."""
        cmd = 'hadoop job -kill %s' % job_id
        hadoop_utils.run_and_check_command_in_hadoop(
            self.hadoop_host_config,
            cmd)

    def run_hadoop_job(self,
                       class_name,
                       jobconf_args=None,
                       extra_args=None,
                       extra_jars=None):
        jobconf_args = jobconf_args if jobconf_args else {}
        extra_args = extra_args if extra_args else []
        extra_jars = extra_jars if extra_jars else []

        # Set default jobconf args
        jobconf_args = jobconf_args.copy()
        if self.config.SCHEDULER_QUEUE:
            jobconf_args[self.config.SCHEDULER_PARAM] = \
                self.config.SCHEDULER_QUEUE
        jobconf_args['mapred.job.name'] = self.job_name

        # create arguments string
        arguments = \
            ' '.join('-D%s=%s' % (k, v) for k, v in jobconf_args.iteritems())
        arguments += ' '
        arguments += ' '.join(extra_args)

        base_dir = self.get_job_resource_dir(self.config.USER)
        libjars_glob = ' '.join(
            ['%s/%s/*.jar' % (base_dir, d) for d in self.config.USER_LIBJAR_DIRS])
        libjars = '`echo %s | tr \' \' \',\'`' % libjars_glob

        user_jar_dirs = ['%s/%s/*' % (base_dir, d) for d in self.config.USER_LIBJAR_DIRS]
        hadoop_classpath = ':'.join(user_jar_dirs)

        app_jar_path = '%s/%s' % (base_dir, self.config.USER_APPJAR_PATH)

        # temp dir for holding stdout and stderr
        query_dir = self._mk_tmp_dir()

        # generate command
        var_dict = {
            'class_name': class_name,
            'arguments': arguments,
            'query_dir': query_dir,
            'app_jar': app_jar_path,
            'libjars': libjars,
            'hadoop_classpath': hadoop_classpath,
        }
        cmd = ('set -o pipefail; HADOOP_CLASSPATH=%(hadoop_classpath)s'
               ' hadoop jar %(app_jar)s'
               ' %(class_name)s'
               ' -libjars %(libjars)s'
               ' %(arguments)s'
               ' 2>&1 > %(query_dir)s/out.csv | tee %(query_dir)s/out.err') % \
            var_dict

        self.log.info('Running class:%s with arguments:%s' % (class_name, arguments))
        self.log.info('Full command: %s' % cmd)

        # run command
        hadoop_utils.run_and_check_command_in_hadoop(
            self.hadoop_host_config,
            cmd,
            log_line_processor=self._hadoop_job_log_line_processor)

        rows = self._get_raw_query_result('%s/out.csv' % query_dir)
        stderr = self._get_raw_query_result('%s/out.err' % query_dir)

        return rows, stderr, self.job_ids

    def _hadoop_job_log_line_processor(self, line):
        """ A callback function that gets executed for every line of
        stderr coming from the running job. Returns a dict of pinball
        metadata.
        """
        job_regex = r"submitted hadoop job: (?P<job_id>.+)"
        m = re.search(job_regex, line)
        if m:
            job_id = m.group('job_id')
            self.job_ids.append(job_id)
            return {'job_id': job_id,
                    'kill_id': '%s/%s' % (self.config.PLATFORM, job_id)}
        return {}

    def get_job_resource_dir(self, run_as_user):
        """Path to location where job resources (jars, archives) are stored.

        For EMR it would be a local dir on the master node."""
        return os.path.join(self.config.HADOOP_HOST_HOME,
                            run_as_user)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Definition of schedule metadata included in schedule tokens.

A Schedule defines when something should run.  (The thing is abstract; a
WorkflowSchedule runs workflows.)  It has a starting run time, and is
(optionally) repeated periodically.  A schedule also has an OverrunPolicy that
defines how it should behave if a previous run didn't finish by the time the
thing is run again through this schedule.
"""
import abc
import datetime
import math
import time

from pinball.config.pinball_config import PinballConfig
from pinball.config.utils import get_log
from pinball.config.utils import timestamp_to_str
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.parser.config_parser import ParserCaller
from pinball.parser.utils import annotate_parser_caller
from pinball.persistence.token_data import TokenData
from pinball.scheduler.overrun_policy import OverrunPolicy
from pinball.ui.data import Status
from pinball.ui.data_builder import DataBuilder
from pinball.workflow.name import Name
from pinball.workflow.signaller import Signal
from pinball.workflow.signaller import Signaller
from pinball.workflow.utils import load_path


__author__ = 'Pawel Garbacki, Mao Ye'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = ['Pawel Garbacki', 'Mao Ye']
__license__ = 'Apache'
__version__ = '2.0'


LOG = get_log('pinball.scheduler.schedule')


class Schedule(TokenData):
    """Parent class for specialized schedule types."""
    __metaclass__ = abc.ABCMeta

    def __init__(self, next_run_time=None, recurrence_seconds=None,
                 overrun_policy=OverrunPolicy.SKIP):
        self.next_run_time = next_run_time
        self.recurrence_seconds = recurrence_seconds
        self.overrun_policy = overrun_policy

    def advance_next_run_time(self):
        """Advance the scheduled run time beyond the current time."""
        now = time.time()
        if self.next_run_time <= now:
            # Set next run time to the lowest timestamp based off
            # recurrence that is greater than the current time.
            delta_runs = math.ceil((now - self.next_run_time) /
                                   self.recurrence_seconds)
            LOG.info('advancing the next run time now %f '
                     'next_run_time %d recurrence_seconds %d delta_runs %f',
                     now, self.next_run_time,
                     self.recurrence_seconds, delta_runs)
            self.next_run_time += int(delta_runs * self.recurrence_seconds)
            if now == self.next_run_time:
                self.next_run_time += self.recurrence_seconds
        assert self.next_run_time > now

    def corresponds_to(self, schedule):
        """Assess correspondence to another schedule.

        Schedules correspond to each other if their next run times are shifted
        by a multiplication of recurrence seconds, and all other fields are
        the same.

        Args:
            schedule: The schedule to compare with.
        Returns:
            True iff the schedules correspond to each other.
        """
        if (self.overrun_policy != schedule.overrun_policy or
                self.recurrence_seconds != schedule.recurrence_seconds):
            return False
        delta = self.next_run_time - schedule.next_run_time
        delta_multiplicator = 1. * delta / self.recurrence_seconds
        return delta_multiplicator == int(delta_multiplicator)

    @abc.abstractmethod
    def run(self, emailer, store):
        """Run the routine pointed to by this schedule."""
        return None

    @abc.abstractmethod
    def is_running(self, store):
        """Checks if the previous run is still active.

        Args:
            store: The store to query for status.
        Returns:
            True iff the run is running.
        """
        return False

    @abc.abstractmethod
    def is_failed(self, store):
        """Checks if the most recent run has failed.

        Args:
            store: The store to query for status.
        Returns:
            True iff the run has failed.
        """
        return False

    @abc.abstractmethod
    def abort_running(self, client, store):
        """Abort all active runs.

        Args:
            client: The client to communicate with the master.
            store: The store to retrieve runs status.
        Returns:
            True iff the workflow has been aborted.
        """
        return False


class WorkflowSchedule(Schedule):
    """Schedule for a workflow."""
    def __init__(
            self,
            next_run_time=None,
            recurrence_seconds=None,
            overrun_policy=OverrunPolicy.SKIP,
            parser_params=PinballConfig.PARSER_PARAMS,
            workflow=None,
            emails=None,
            max_running_instances=None):
        Schedule.__init__(self, next_run_time, recurrence_seconds,
                          overrun_policy)
        self.parser_params = annotate_parser_caller(
            parser_params,
            ParserCaller.SCHEDULE)

        self.workflow = workflow
        self.emails = emails if emails is not None else []
        self.max_running_instances = max_running_instances if max_running_instances \
            else PinballConfig.DEFAULT_MAX_WORKFLOW_RUNNING_INSTANCES

    def __str__(self):
        if self.next_run_time:
            next_run_time = timestamp_to_str(self.next_run_time)
        else:
            next_run_time = str(self.next_run_time)
        if self.recurrence_seconds:
            delta = datetime.timedelta(seconds=self.recurrence_seconds)
            recurrence = str(delta)
        else:
            recurrence = str(self.recurrence_seconds)
        if self.overrun_policy is not None:
            overrun_policy = OverrunPolicy.to_string(self.overrun_policy)
        else:
            overrun_policy = str(self.overrun_policy)
        return ('WorkflowSchedule(next_run_time=%s, recurrence=%s, '
                'overrun_policy=%s, parser_params=%s, workflow=%s, '
                'email=%s, max_running_instances=%s)' % (next_run_time,
                                                         recurrence,
                                                         overrun_policy,
                                                         self.parser_params,
                                                         self.workflow,
                                                         self.emails,
                                                         str(self.max_running_instances)))

    def __repr__(self):
        return self.__str__()

    def advance_next_run_time(self):
        # TODO(pawel): remove after debugging.
        LOG.info('advancing the next run time for workflow %s', self.workflow)
        super(WorkflowSchedule, self).advance_next_run_time()

    def corresponds_to(self, schedule):
        if (self.parser_params != schedule.parser_params or
                self.workflow != schedule.workflow or
                self.emails != schedule.emails or
                self.max_running_instances != schedule.max_running_instances):
            return False
        return super(WorkflowSchedule, self).corresponds_to(schedule)

    def run(self, emailer, store):
        if not self._check_workflow_instances(emailer, self.workflow, store):
            LOG.warn('too many instances running for workflow %s', self.workflow)
            return None

        config_parser = load_path(PinballConfig.PARSER)(self.parser_params)
        workflow_tokens = config_parser.get_workflow_tokens(self.workflow)
        if not workflow_tokens:
            LOG.error('workflow %s not found', self.workflow)
            return None
        result = ModifyRequest()
        result.updates = workflow_tokens
        assert result.updates
        token = result.updates[0]
        name = Name.from_job_token_name(token.name)
        if not name.instance:
            name = Name.from_event_token_name(token.name)
        LOG.info('exporting workflow %s instance %s.  Its tokens are under %s',
                 name.workflow, name.instance, name.get_instance_prefix())
        return result

    def is_running(self, store):
        data_builder = DataBuilder(store, use_cache=True)
        workflow_data = data_builder.get_workflow(self.workflow)
        if not workflow_data:
            return False
        return workflow_data.status == Status.RUNNING

    def is_failed(self, store):
        data_builder = DataBuilder(store, use_cache=True)
        workflow_data = data_builder.get_workflow(self.workflow)
        if not workflow_data:
            return False
        return (workflow_data.status != Status.RUNNING and
                workflow_data.status != Status.SUCCESS)

    def _get_running_instances(self, store):
        """Find running instances of the workflow.

        Args:
            store: The store to query for wokflow instance status.
        Returns:
            List of running workflow instance names.
        """
        data_builder = DataBuilder(store, use_cache=True)
        instances = data_builder.get_instances(self.workflow)
        result = []
        for instance in instances:
            if instance.status == Status.RUNNING:
                result.append(instance.instance)
        return result

    def abort_running(self, client, store):
        running_instances = self._get_running_instances(store)
        for instance in running_instances:
            signaller = Signaller(client,
                                  workflow=self.workflow,
                                  instance=instance)
            signaller.set_action(Signal.ABORT)
            if not signaller.is_action_set(Signal.ABORT):
                return False
        return True

    def _check_workflow_instances(self, emailer, workflow_name, store):
        """Check the number of running instances of the workflow.

        Besides of the return, also send out warning email if too many
        instances running for the given workflow.

        Args:
            emailer: The email sender.
            workflow_name: Name of the workflow.
            store: The store to retrieve runs status.

        Returns:
            False if running instance number exceeds the max_running_instances setting;
            Otherwise, True.
        """
        running_instances = self._get_running_instances(store)
        if self.max_running_instances and len(running_instances) >= self.max_running_instances:
            LOG.warn('Too many (%s) instances running for workflow %s !',
                     len(running_instances), workflow_name)
            if emailer:
                emailer.send_too_many_running_instances_warning_message(self.emails,
                                                                        workflow_name,
                                                                        len(running_instances),
                                                                        self.max_running_instances)
            else:
                LOG.warn('Emailer is not set! Failed to send too many instances running warning '
                         'email for workflow %s', workflow_name)
            return False
        return True



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for operations on token names."""
import unittest

from pinball.workflow.name import Name


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class NameTestCase(unittest.TestCase):
    def test_workflow_prefix(self):
        PREFIX = '/workflow/some_workflow/'
        name = Name.from_workflow_prefix(PREFIX)
        self.assertEqual('some_workflow', name.workflow)
        self.assertEqual(PREFIX, name.get_workflow_prefix())

    def test_instance_prefix(self):
        PREFIX = '/workflow/some_workflow/some_instance/'
        name = Name.from_instance_prefix(PREFIX)
        self.assertEqual('some_workflow', name.workflow)
        self.assertEqual('some_instance', name.instance)
        self.assertEqual(PREFIX, name.get_instance_prefix())

    def test_job_state_prefix(self):
        PREFIX = '/workflow/some_workflow/some_instance/job/waiting/'
        name = Name.from_job_state_prefix(PREFIX)
        self.assertEqual('some_workflow', name.workflow)
        self.assertEqual('some_instance', name.instance)
        self.assertEqual('waiting', name.job_state)
        self.assertEqual(PREFIX, name.get_job_state_prefix())

    def test_job_prefix(self):
        PREFIX = '/workflow/some_workflow/some_instance/job/'
        name = Name.from_job_prefix(PREFIX)
        self.assertEqual('some_workflow', name.workflow)
        self.assertEqual('some_instance', name.instance)
        self.assertEqual(PREFIX, name.get_job_prefix())

    def test_job_token_name(self):
        NAME = '/workflow/some_workflow/some_instance/job/waiting/some_job'
        name = Name.from_job_token_name(NAME)
        self.assertEqual('some_workflow', name.workflow)
        self.assertEqual('some_instance', name.instance)
        self.assertEqual('waiting', name.job_state)
        self.assertEqual('some_job', name.job)
        self.assertEqual(NAME, name.get_job_token_name())

    def test_input_prefix(self):
        PREFIX = ('/workflow/some_workflow/some_instance/input/some_job/'
                  'some_input/')
        name = Name.from_input_prefix(PREFIX)
        self.assertEqual('some_workflow', name.workflow)
        self.assertEqual('some_instance', name.instance)
        self.assertEqual('some_job', name.job)
        self.assertEqual('some_input', name.input)
        self.assertEqual(PREFIX, name.get_input_prefix())

    def test_event_token_name(self):
        NAME = ('/workflow/some_workflow/some_instance/input/some_job/'
                'some_input/some_event')
        name = Name.from_event_token_name(NAME)
        self.assertEqual('some_workflow', name.workflow)
        self.assertEqual('some_instance', name.instance)
        self.assertEqual('some_job', name.job)
        self.assertEqual('some_input', name.input)
        self.assertEqual('some_event', name.event)
        self.assertEqual(NAME, name.get_event_token_name())

    def test_workflow_schedule_token_name(self):
        NAME = '/schedule/workflow/some_workflow'
        name = Name.from_workflow_schedule_token_name(NAME)
        self.assertEqual('some_workflow', name.workflow)
        self.assertEqual(NAME, name.get_workflow_schedule_token_name())

    def test_signal_prefix(self):
        TOP_SIGNAL_PREFIX = '/workflow/__SIGNAL__/'
        name = Name.from_signal_token_name(TOP_SIGNAL_PREFIX)
        self.assertEqual(TOP_SIGNAL_PREFIX, name.get_signal_prefix())

        WORKFLOW_SIGNAL_PREFIX = '/workflow/some_workflow/__SIGNAL__/'
        name.workflow = 'some_workflow'
        self.assertEqual(WORKFLOW_SIGNAL_PREFIX, name.get_signal_prefix())

        INSTANCE_SIGNAL_PREFIX = ('/workflow/some_workflow/some_instance/'
                                  '__SIGNAL__/')
        name.instance = 'some_instance'
        self.assertEqual(INSTANCE_SIGNAL_PREFIX, name.get_signal_prefix())

    def test_signal_token_name(self):
        TOP_SIGNAL_NAME = '/workflow/__SIGNAL__/some_signal'
        name = Name.from_signal_token_name(TOP_SIGNAL_NAME)
        self.assertEqual('some_signal', name.signal)
        self.assertIsNone(name.workflow)
        self.assertEqual(TOP_SIGNAL_NAME, name.get_signal_token_name())

        WORKFLOW_SIGNAL_NAME = ('/workflow/some_workflow/__SIGNAL__/'
                                'some_signal')
        name = Name.from_signal_token_name(WORKFLOW_SIGNAL_NAME)
        self.assertEqual('some_signal', name.signal)
        self.assertEqual('some_workflow', name.workflow)
        self.assertIsNone(name.instance)
        self.assertEqual(WORKFLOW_SIGNAL_NAME, name.get_signal_token_name())

        INSTANCE_SIGNAL_NAME = ('/workflow/some_workflow/some_instance/'
                                '__SIGNAL__/some_signal')
        name = Name.from_signal_token_name(INSTANCE_SIGNAL_NAME)
        self.assertEqual('some_signal', name.signal)
        self.assertEqual('some_workflow', name.workflow)
        self.assertEqual('some_instance', name.instance)
        self.assertEqual(INSTANCE_SIGNAL_NAME, name.get_signal_token_name())



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for the analyzer."""
import collections
import pickle
import unittest

from pinball.master.thrift_lib.ttypes import Token
from pinball.workflow.analyzer import Analyzer
from pinball.workflow.event import Event
from pinball.workflow.job import ShellJob
from pinball.workflow.job_executor import ExecutionRecord
from pinball.workflow.name import Name
from tests.pinball.persistence.ephemeral_store import EphemeralStore


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class AnalyzerTestCase(unittest.TestCase):
    _NUM_LEVELS = 3

    def setUp(self):
        self._store = EphemeralStore()

    def _add_active_workflow_tokens(self):
        """Add some active workflow tokens.

        The job dependencies form a complete binary tree turned upside down.
        I.e., each job has two parents.
        """
        self._store = EphemeralStore()
        version = 1
        for level in range(AnalyzerTestCase._NUM_LEVELS):
            jobs_at_level = 2 ** (AnalyzerTestCase._NUM_LEVELS - level - 1)
            for job_index in range(jobs_at_level):
                job_name = 'job_%d_%d' % (level, job_index)
                event_name = Name(workflow='some_workflow',
                                  instance='123',
                                  job=job_name,
                                  event='some_event')
                if level == 0:
                    inputs = [Name.WORKFLOW_START_INPUT,
                              Name.WORKFLOW_START_INPUT + '_prime']
                    event_name.input = Name.WORKFLOW_START_INPUT
                else:
                    inputs = ['job_%d_%d' % (level - 1, 2 * job_index),
                              'job_%d_%d' % (level - 1, 2 * job_index + 1)]
                    event_name.input = 'job_%d_%d' % (level - 1, 2 * job_index)
                if level == AnalyzerTestCase._NUM_LEVELS - 1:
                    outputs = []
                else:
                    outputs = ['job_%d_%d' % (level + 1, job_index / 2)]
                job = ShellJob(name=job_name,
                               inputs=inputs,
                               outputs=outputs,
                               command='some_command')
                job.history.append(ExecutionRecord())
                name = Name(workflow='some_workflow', instance='123',
                            job_state=Name.WAITING_STATE, job=job_name)
                job_token = Token(version=version,
                                  name=name.get_job_token_name(),
                                  priority=10,
                                  data=pickle.dumps(job))
                version += 1
                event = Event('some_event')
                event_token = Token(version=version,
                                    name=event_name.get_event_token_name(),
                                    priority=10,
                                    data=pickle.dumps(event))
                self._store.commit_tokens([job_token, event_token])

    def _archive_tokens(self):
        tokens = self._store.read_active_tokens()
        self._store.archive_tokens(tokens)
        return tokens

    def _simulate(self):
        """Simulate execution of active jobs."""
        tokens = self._store.read_tokens()
        satisfied_deps = set()
        executed_jobs = []
        jobs = {}
        for token in tokens:
            event_name = Name.from_event_token_name(token.name)
            if event_name.event:
                satisfied_deps.add((event_name.input, event_name.job))
            else:
                job_name = Name.from_job_token_name(token.name)
                if job_name.job:
                    job = pickle.loads(token.data)
                    jobs[job.name] = job
        dep_counts = collections.defaultdict(int)
        while satisfied_deps:
            last_satisfied_deps = satisfied_deps
            satisfied_deps = set()
            for (_, job_name) in last_satisfied_deps:
                dep_counts[job_name] += 1
                if dep_counts[job_name] == 2:
                    executed_jobs.append(job_name)
                    job = jobs[job_name]
                    for output in job.outputs:
                        satisfied_deps.add((job_name, output))
        return executed_jobs

    def test_change_instance(self):
        self._add_active_workflow_tokens()
        self._archive_tokens()
        analyzer = Analyzer.from_store(self._store, 'some_workflow', '123')
        analyzer.change_instance('321')
        tokens = analyzer.get_tokens()
        self.assertLess(0, len(tokens))
        for token in tokens:
            name = Name.from_job_token_name(token.name)
            self.assertEqual('321', name.instance)

    def test_change_job_histories(self):
        self._add_active_workflow_tokens()
        self._archive_tokens()
        analyzer = Analyzer.from_store(self._store, 'some_workflow', '123')
        analyzer.clear_job_histories()
        tokens = analyzer.get_tokens()
        self.assertLess(0, len(tokens))
        for token in tokens:
            job = pickle.loads(token.data)
            self.assertEqual([], job.history)

    def test_poison_no_tokens(self):
        analyzer = Analyzer.from_store(self._store, 'some_workflow', '123')
        analyzer.poison([])

    def test_poison_no_roots(self):
        self._add_active_workflow_tokens()
        self._archive_tokens()
        analyzer = Analyzer.from_store(self._store, 'some_workflow', '123')
        analyzer.poison([])
        tokens = analyzer.get_tokens()
        self._store.commit_tokens(updates=tokens)
        executed_jobs = self._simulate()
        self.assertEqual([], executed_jobs)

    def test_poison_all(self):
        """Poison all top level jobs."""
        self._add_active_workflow_tokens()
        self._archive_tokens()
        analyzer = Analyzer.from_store(self._store, 'some_workflow', '123')

        roots = []
        for job_index in range(0, 2 ** (AnalyzerTestCase._NUM_LEVELS - 1)):
            roots.append('job_0_%d' % job_index)

        analyzer.poison(roots)
        tokens = analyzer.get_tokens()
        self._store.commit_tokens(updates=tokens)
        executed_jobs = self._simulate()
        # We expect that every job has run.
        expected_num_executed_jobs = 2 ** (AnalyzerTestCase._NUM_LEVELS) - 1
        self.assertEqual(expected_num_executed_jobs, len(executed_jobs))

    def test_poison_subset(self):
        """Poison every second top level job."""
        self._add_active_workflow_tokens()
        self._archive_tokens()
        analyzer = Analyzer.from_store(self._store, 'some_workflow', '123')

        roots = []
        for job_index in range(0, 2 ** (AnalyzerTestCase._NUM_LEVELS - 1), 2):
            roots.append('job_0_%d' % job_index)

        analyzer.poison(roots)
        tokens = analyzer.get_tokens()
        self._store.commit_tokens(updates=tokens)
        executed_jobs = self._simulate()
        # We expect that every second job at the top level and every job at
        # a lower level was run.
        expected_num_executed_jobs = (
            2 ** (AnalyzerTestCase._NUM_LEVELS - 1) - 1 +
            2 ** (AnalyzerTestCase._NUM_LEVELS - 1) / 2)
        self.assertEqual(expected_num_executed_jobs, len(executed_jobs))

    def test_poison_get_new_event_tokens(self):
        """Poison all top level jobs and get new event tokens."""
        self._add_active_workflow_tokens()
        tokens = self._archive_tokens()
        analyzer = Analyzer.from_store(self._store, 'some_workflow', '123')
        analyzer._filter_event_tokens(tokens)

        roots = []
        for job_index in range(0, 2 ** (AnalyzerTestCase._NUM_LEVELS - 1)):
            roots.append('job_0_%d' % job_index)

        analyzer.poison(roots)
        tokens = analyzer.get_new_event_tokens()
        expected_num_new_event_tokens = 2 ** AnalyzerTestCase._NUM_LEVELS
        self.assertEqual(expected_num_new_event_tokens, len(tokens))



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for test cases."""

import StringIO


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


def setup(setup_method):
    """Decorator for declaring setup methods for test cases."""
    setup_method._is_setup_method = True
    return setup_method


def teardown(teardown_method):
    """Decorator for declaring teardown methods for test cases."""
    teardown_method._is_teardown_method = True
    return teardown_method


class StringIOWithContext(StringIO.StringIO):
    """Subclass of StringIO to allow it to be used as a context manager."""
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_t):
        # Don't suppress exceptions!
        return False



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An executor that executes Hadoop job locally."""

import os

from pinball_ext.executor.cluster_executor import ClusterExecutor
from pinball_ext.executor.common import Platform


__author__ = 'Changshu Liu, Rui Jiang'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class LocalExecutor(ClusterExecutor):
    """Runs Hadoop jobs on local machine.

    The Config.USER_LIBJAR_DIRS, Config.USER_APP_JAR_PATH should be local path.
    """
    def __init__(self, executor_config=None):
        super(LocalExecutor, self).__init__(executor_config)
        self.config.PLATFORM = Platform.LOCAL

    def run_hadoop_job(self,
                       class_name,
                       jobconf_args=None,
                       extra_args=None,
                       extra_jars=None):
        jobconf_args = jobconf_args if jobconf_args else {}
        extra_args = extra_args if extra_args else []

        # set default jobconf args
        jobconf_args = jobconf_args.copy()
        jobconf_args['mapred.job.name'] = self.job_name

        # create arguments string
        arguments = ' '.join('-D%s=%s' % (k, v) for k, v in jobconf_args.iteritems())
        arguments += ' '
        arguments += ' '.join(extra_args)

        libjars_glob = ' '.join(
            ['%s/*.jar' % d for d in self.config.USER_LIBJAR_DIRS])
        libjars = '`echo %s | tr \' \' \',\'`' % libjars_glob

        user_jar_dirs = ['%s/*' % d for d in self.config.USER_LIBJAR_DIRS]
        hadoop_classpath = ':'.join(user_jar_dirs)

        # generate command
        var_dict = {
            'user': self.config.USER,
            'class_name': class_name,
            'arguments': arguments,
            'hadoop_classpath': hadoop_classpath,
            'libjars': libjars,
            'app_jar': self.config.USER_APPJAR_PATH,
        }

        cmd = ('export HADOOP_CLASSPATH=%(hadoop_classpath)s;'
               ' hadoop jar %(app_jar)s'
               ' %(class_name)s'
               ' -libjars %(libjars)s'
               ' %(arguments)s') % var_dict

        # log command message
        self.log.info('Running class:%s with arguments:%s' %
                      (class_name, arguments))
        self.log.info('Full command: %s' % cmd)

        # run command
        os.system(cmd)

        return [], [], self.job_ids



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Various utilities related to Hive."""

import re


__author__ = 'Zach Drach, Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


def extract_target_tables(hive_query):
    """Extracts the names of the tables that were updated.

    Returns:
        A list of (database,table) names that were updated.
    """
    flags = re.MULTILINE | re.IGNORECASE | re.DOTALL
    pattern = r"\b(?:insert\s+overwrite\s+table\s+'?(?:(\w+)\.)?" \
              r"(\w+)'?\s+partition)|(?:use\s+'?(\w+)'?\s*;)"
    current_database = 'default'
    output = []
    for match in re.finditer(pattern,
                             strip_hive_comments(hive_query),
                             flags):
        groups = match.groups()
        if groups[2] is not None:
            current_database = groups[2]
        else:
            database = groups[0] if groups[0] else current_database
            table = groups[1]
            output.append((database, table))
    return output


def strip_hive_comments(hive_query):
    """Strip the comments in a Hive query."""
    regex = r'--.*'
    flags = re.MULTILINE | re.IGNORECASE
    return re.sub(regex, '', hive_query, flags)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Event signals system state change.

Events are posted to job inputs.  Job input is a bucket for tokens representing
a class of events (e.g., completion of a specific upstream job).  A job becomes
runnable if it has at least one event in each input.  We call a group of
events with one token in each job input 'triggering' events.  Triggering events
get consumed (i.e., removed from the master) when the job becomes runnable.
"""
from pinball.persistence.token_data import TokenData


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class Event(TokenData):
    def __init__(self, creator=None, attributes=None):
        self.creator = creator
        self.attributes = {} if attributes is None else attributes

    @property
    def _COMPATIBILITY_ATTRIBUTES(self):
        return {
            'attributes': {}
        }

    def __str__(self):
        return 'Event(creator=%s, attributes=%s)' % (self.creator,
                                                     self.attributes)

    def __repr__(self):
        return self.__str__()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
import unittest

from pinball_ext.executor import local_executor


__author__ = 'Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class LocalExecutorTest(unittest.TestCase):
    @mock.patch('os.system')
    def test_hadoop_job(self, f1):
        user_jar_dirs = ['/dir1/jar1', '/dir2/jar2']
        user_app_jar = '/dir/jar.jar'
        user_archive = '/dir/file.archive'

        executor = local_executor.LocalExecutor(
            executor_config={
                'USER_LIBJAR_DIRS': ','.join(user_jar_dirs),
                'USER_APPJAR_PATH': user_app_jar,
                'USER_ARCHIVE_PATH': user_archive,
                'USER': 'test_user'
            }
        )
        executor.run_hadoop_job(
            'test_job_class',
            jobconf_args={
                'pinball.key1': 'value1',
                'pinball.key2': 'value2',
            },
            extra_args=[
                '-Dpinball.key3=value3'
            ])

        self.assertEqual(executor.config.USER_LIBJAR_DIRS, user_jar_dirs)
        self.assertEqual(executor.config.USER_APPJAR_PATH, user_app_jar)
        self.assertEqual(executor.config.USER_ARCHIVE_PATH, user_archive)
        # self.assertEqual(executor.config.PLATFORM, Platform.LOCAL)
        f1.assert_called_once_with(
            "export HADOOP_CLASSPATH=/dir1/jar1/*:/dir2/jar2/*; "
            "hadoop jar /dir/jar.jar test_job_class "
            "-libjars `echo /dir1/jar1/*.jar /dir2/jar2/*.jar | tr ' ' ','` "
            "-Dmapred.job.name=test_user:AdHocCommand "
            "-Dpinball.key2=value2 "
            "-Dpinball.key1=value1 "
            "-Dpinball.key3=value3")



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for data builder."""
import mock
import sys
import unittest

from pinball.config.utils import PinballException
from pinball.ui.data import Status
from pinball.ui.data import WorkflowInstanceData
from pinball.ui.data_builder import DataBuilder
from pinball.workflow.signaller import Signal
from tests.pinball.persistence.ephemeral_store import EphemeralStore
from tests.pinball.persistence.data_generator import \
    generate_workflows


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class DataBuilderTestCase(unittest.TestCase):
    def setUp(self):
        self._store = EphemeralStore()
        self._data_builder = DataBuilder(self._store)

    @mock.patch('os.makedirs')
    @mock.patch('__builtin__.open')
    def _add_tokens(self, _, __):
        generate_workflows(2, 2, 2, 2, 2, self._store)

    def test_get_workflows_empty(self):
        self.assertEqual([], self._data_builder.get_workflows())

    def _get_workflows(self):
        self._add_tokens()
        workflows = self._data_builder.get_workflows()
        self.assertEqual(4, len(workflows))
        workflow_status = {'workflow_0': Status.RUNNING,
                           'workflow_1': Status.RUNNING,
                           'workflow_2': Status.SUCCESS,
                           'workflow_3': Status.FAILURE}
        for workflow in workflows:
            self.assertEqual(workflow_status[workflow.workflow],
                             workflow.status)
            self.assertEqual('instance_1', workflow.last_instance)
            del workflow_status[workflow.workflow]
        self.assertEqual({}, workflow_status)

    def test_get_workflows(self):
        self._get_workflows()

    def test_get_workflows_using_cache(self):
        self._data_builder.use_cache = True
        self._get_workflows()
        # Only finished (archived) workflow instances should have been cached.
        expected_cached_names = ['/workflow/workflow_2/instance_0/',
                                 '/workflow/workflow_2/instance_1/',
                                 '/workflow/workflow_3/instance_0/',
                                 '/workflow/workflow_3/instance_1/']
        cached_names = sorted(self._store.read_cached_data_names())
        self.assertEqual(expected_cached_names, cached_names)

    def test_get_workflow_empty(self):
        self.assertIsNone(self._data_builder.get_workflow('does_not_exist'))

    def _get_workflow(self):
        self._add_tokens()
        workflow = self._data_builder.get_workflow('workflow_0')
        self.assertEqual('workflow_0', workflow.workflow)
        self.assertEqual(Status.RUNNING, workflow.status)
        self.assertEqual('instance_1', workflow.last_instance)

    def test_get_workflow(self):
        self._get_workflow()

    def test_get_workflow_using_cache(self):
        self._data_builder.use_cache = True
        self._get_workflow()
        # Instances of a running workflow should not have been cached.
        self.assertEqual([], self._store.read_cached_data_names())

    def test_get_instances_empty(self):
        self.assertEqual([],
                         self._data_builder.get_instances('does_not_exist'))

    def _get_instances(self):
        self._add_tokens()
        instances = self._data_builder.get_instances('workflow_2')
        self.assertEqual(2, len(instances))
        instance_status = [Status.SUCCESS, Status.FAILURE]
        for instance in instances:
            self.assertEqual('workflow_2', instance.workflow)
            instance_status.remove(instance.status)
        self.assertEqual([], instance_status)

    def test_get_instances(self):
        self._get_instances()

    def test_get_instances_using_cache(self):
        self._data_builder.use_cache = True
        self._get_instances()
        expected_cached_names = ['/workflow/workflow_2/instance_0/',
                                 '/workflow/workflow_2/instance_1/']
        cached_names = sorted(self._store.read_cached_data_names())
        self.assertEqual(expected_cached_names, cached_names)

    def test_get_instance_empty(self):
        self.assertIsNone(None,
                          self._data_builder.get_instance('does_not_exist',
                                                          'instance_0'))

    def _get_instance(self):
        self._add_tokens()
        instance = self._data_builder.get_instance('workflow_0', 'instance_0')
        self.assertEqual('workflow_0', instance.workflow)
        self.assertEqual('instance_0', instance.instance)

    def test_get_instance(self):
        self._get_instance()

    def test_get_instance_using_cache(self):
        self._data_builder.use_cache = True
        self._get_instance()
        # Running instance should not have been cached.
        self.assertEqual([], self._store.read_cached_data_names())

    def test_get_jobs_empty(self):
        self.assertEqual([],
                         self._data_builder.get_jobs('does_not_exist',
                                                     'does_not_exist'))

    def test_get_jobs(self):
        self._add_tokens()
        jobs = self._data_builder.get_jobs('workflow_0', 'instance_0')
        self.assertEqual(2, len(jobs))
        for job in jobs:
            self.assertEqual('workflow_0', job.workflow)
            self.assertEqual('instance_0', job.instance)
            self.assertEqual('ShellJob', job.job_type)
            self.assertTrue(job.info.startswith('command=some command'))
            self.assertEqual(Status.FAILURE, job.status)
        self.assertEqual([(0, ''), (1, 'SUCCESS'), (9, 'FAILURE')],
                         jobs[0].progress)
        self.assertEqual([(89, ''), (1, 'SUCCESS'), (9, 'FAILURE')],
                         jobs[1].progress)

    def test_get_executions_empty(self):
        self.assertEqual([],
                         self._data_builder.get_executions('does_not_exist',
                                                           'does_not_exist',
                                                           'does_not_exist'))

    def test_get_executions(self):
        self._add_tokens()
        executions = self._data_builder.get_executions('workflow_0',
                                                       'instance_0',
                                                       'job_0')
        self.assertEqual(2, len(executions))
        exit_codes = [0, 1]
        for execution in executions:
            self.assertEqual('workflow_0', execution.workflow)
            self.assertEqual('instance_0', execution.instance)
            self.assertEqual('job_0', execution.job)
            self.assertTrue(execution.info.startswith('some_command'))
            exit_codes.remove(execution.exit_code)
            self.assertEqual(2, len(execution.logs))

    def test_get_executions_across_instances_empty(self):
        self.assertEqual([],
                         self._data_builder.get_executions_across_instances(
                             'does_not_exist',
                             'does_not_exist'))

    def test_get_executions_across_instances(self):
        self._add_tokens()
        executions = self._data_builder.get_executions_across_instances(
            'workflow_0', 'job_0')
        self.assertEqual(2 * 2, len(executions))
        exit_codes = [0, 0, 1, 1]
        for execution in executions:
            self.assertEqual('workflow_0', execution.workflow)
            self.assertEqual('job_0', execution.job)
            self.assertTrue(execution.info.startswith('some_command'))
            exit_codes.remove(execution.exit_code)
            self.assertEqual(2, len(execution.logs))

    def test_get_execution_empty(self):
        self.assertIsNone(self._data_builder.get_execution('does_not_exist',
                                                           'does_not_exist',
                                                           'does_not_exist',
                                                           0))

    def test_get_execution(self):
        self._add_tokens()
        execution = self._data_builder.get_execution('workflow_0',
                                                     'instance_0',
                                                     'job_0',
                                                     1)
        self.assertEqual('workflow_0', execution.workflow)
        self.assertEqual('instance_0', execution.instance)
        self.assertEqual('job_0', execution.job)
        self.assertEqual(1, execution.execution)
        self.assertEqual('some_command 1 some_args 1', execution.info)
        self.assertEqual(1, execution.exit_code)
        self.assertEqual(2, execution.start_time)
        self.assertEqual(13, execution.end_time)
        self.assertEqual(2, len(execution.logs))

    @mock.patch('__builtin__.open')
    def test_get_file_content_no_file(self, _):
        self.assertEqual('',
                         self._data_builder.get_file_content('does_not_exist',
                                                             'does_not_exist',
                                                             'does_not_exist',
                                                             'does_not_exist',
                                                             'does_not_exist'))

    @mock.patch('os.makedirs')
    @mock.patch('__builtin__.open')
    def test_get_file_content(self, open_mock, _):
        generate_workflows(2, 2, 2, 2, 2, self._store)

        file_mock = mock.MagicMock()
        open_mock.return_value = file_mock
        file_mock.__enter__.return_value = file_mock
        file_mock.read.return_value = 'some_content'

        content = self._data_builder.get_file_content('workflow_0',
                                                      'instance_0',
                                                      'job_0',
                                                      0,
                                                      'info')
        self.assertEqual('some_content', content)

    def test_get_token_paths_empty(self):
        self.assertRaises(PinballException,
                          self._data_builder.get_token_paths,
                          '')

    def test_get_token_paths(self):
        self._add_tokens()
        token_paths = self._data_builder.get_token_paths(
            '/workflow/workflow_0/instance_0/job/waiting/')
        self.assertEqual(2, len(token_paths))
        paths = ['/workflow/workflow_0/instance_0/job/waiting/job_0',
                 '/workflow/workflow_0/instance_0/job/waiting/job_1']
        for token_path in token_paths:
            self.assertEqual(1, token_path.count)
            paths.remove(token_path.path)
        self.assertEqual([], paths)

    def test_get_token_empty(self):
        self.assertRaises(PinballException,
                          self._data_builder.get_token,
                          '/does_not_exist')

    def test_get_token(self):
        self._add_tokens()
        token = self._data_builder.get_token(
            '/workflow/workflow_0/instance_0/job/waiting/job_0')
        self.assertEqual('/workflow/workflow_0/instance_0/job/waiting/job_0',
                         token.name)
        self.assertIsNone(token.owner)
        self.assertIsNone(token.expiration_time)
        self.assertEqual(0, token.priority)
        self.assertIsNotNone(token.data)

    def test_signal_not_set(self):
        self.assertFalse(self._data_builder.is_signal_set('does_not_exist', 0,
                                                          Signal.DRAIN))

    def test_signal_set(self):
        self._add_tokens()
        self.assertTrue(self._data_builder.is_signal_set('workflow_0', 0,
                                                         Signal.DRAIN))

    # Workflow status should be the running instance
    def test_workflow_data_from_instances_data1(self):
        wf_instance_list = [
            WorkflowInstanceData('wf', '22346', Status.ABORTED, 12346, 54321),
            WorkflowInstanceData('wf', '22345', Status.RUNNING, 12345, None),
            WorkflowInstanceData('wf', '22347', Status.SUCCESS, 12347, 12390),
            WorkflowInstanceData('wf', '22348', Status.RUNNING, 12348, None),
        ]
        wf_data = DataBuilder._workflow_data_from_instances_data(
            wf_instance_list)
        self.assertEquals(wf_data.workflow, 'wf')
        self.assertEquals(wf_data.status, Status.RUNNING)
        self.assertEquals(wf_data.last_instance, '22348')
        self.assertEquals(wf_data.last_start_time, 12348)
        self.assertEquals(wf_data.last_end_time, None)
        self.assertEquals(wf_data.running_instance_number, 2)

    # Workflow status should be the running instance even if some instance ended
    # at sys.maxint time
    def test_workflow_data_from_instances_data2(self):
        wf_instance_list = [
            WorkflowInstanceData('wf', '22346', Status.ABORTED, 12355, sys.maxint),
            WorkflowInstanceData('wf', '22347', Status.SUCCESS, 12365, 12390),
            WorkflowInstanceData('wf', '22345', Status.RUNNING, 12345, None),
        ]
        wf_data = DataBuilder._workflow_data_from_instances_data(
            wf_instance_list)
        self.assertEquals(wf_data.workflow, 'wf')
        self.assertEquals(wf_data.status, Status.RUNNING)
        self.assertEquals(wf_data.last_instance, '22345')
        self.assertEquals(wf_data.last_start_time, 12345)
        self.assertEquals(wf_data.last_end_time, None)
        self.assertEquals(wf_data.running_instance_number, 1)

    # Workflow status should be the last finished instance
    def test_workflow_data_from_instances_data3(self):
        wf_instance_list = [
            WorkflowInstanceData('wf', '22346', Status.ABORTED, 12345, 12392),
            WorkflowInstanceData('wf', '22347', Status.SUCCESS, 12346, 12393),
            WorkflowInstanceData('wf', '22345', Status.FAILURE, 12347, 12391),
            ]
        wf_data = DataBuilder._workflow_data_from_instances_data(
            wf_instance_list)
        self.assertEquals(wf_data.workflow, 'wf')
        self.assertEquals(wf_data.status, Status.SUCCESS)
        self.assertEquals(wf_data.last_instance, '22347')
        self.assertEquals(wf_data.last_start_time, 12346)
        self.assertEquals(wf_data.last_end_time, 12393)
        self.assertEquals(wf_data.running_instance_number, 0)

    # Workflow status should be the last finished instance even if some instance
    # ended with sys.maxint time
    def test_workflow_data_from_instances_data4(self):
        wf_instance_list = [
            WorkflowInstanceData('wf', '22346', Status.ABORTED, 12345, 12392),
            WorkflowInstanceData('wf', '22347', Status.SUCCESS, 12346, 12393),
            WorkflowInstanceData('wf', '22345', Status.FAILURE, 12391, sys.maxint),
            ]
        wf_data = DataBuilder._workflow_data_from_instances_data(
            wf_instance_list)
        self.assertEquals(wf_data.workflow, 'wf')
        self.assertEquals(wf_data.status, Status.SUCCESS)
        self.assertEquals(wf_data.last_instance, '22347')
        self.assertEquals(wf_data.last_start_time, 12346)
        self.assertEquals(wf_data.last_end_time, 12393)
        self.assertEquals(wf_data.running_instance_number, 0)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class OutputFilter(object):
    """A filter processing a stream and generating magic properties."""
    def __init__(self, log_line_processor, output=sys.stdout):
        self._log_line_processor = log_line_processor
        self._output = output
        self._buffer = None

    def _generate_magic(self, line):
        """Process log line potentially generating Pinball magic properties.

        Args:
            line: The line to process.

        Returns:
            A string with Pinball magic properties or an empty string if no
            properties were generated.
        """
        key_values = self._log_line_processor(line)
        result = ''
        for key, value in key_values.items():
            result += 'PINBALL:%s=%s\n' % (key, value)
        return result

    def _write(self, line):
        """Write a line to the output or to the buffer.

        Args:
            line: The line to write.
        """
        if self._buffer is not None:
            self._buffer += line
        else:
            self._output.write(line)

    def _process_line(self, line):
        """Process a line.

        Args:
            line: The line to process.
        """
        self._write(line)
        magic = self._generate_magic(line)
        if magic:
            if self._buffer is None and not line.endswith('\n'):
                self._write('\n')
            self._output.write(magic)
        self._output.flush()

    def _read_from_input(self, input_stream):
        """Read and process all lines from an input stream.

        Args:
            input_stream: The input stream to read from.
        """
        self._line = ''
        # Iterating directly over input_stream incurs delays in output
        # propagation.
        for line in iter(input_stream.readline, ''):
            self._process_line(line)

    def read_and_output(self, input_stream):
        """Read all lines from an input stream and write to output stream.

        Args:
            input_stream: The input stream to read from.
        """
        self._buffer = None
        self._read_from_input(input_stream)

    def read_and_buffer(self, input_stream):
        """Read all lines from an input stream and write to the buffer.

        Pinball magic properties will be still written to the output rather
        than to the buffer.

        Args:
            input_stream: The input stream to read from.

        Returns:
            The buffer with read lines.
        """
        self._buffer = ''
        self._read_from_input(input_stream)
        return self._buffer

    def process_and_output(self, input_lines):
        """ Reads an iterable of lines and writes to output stream.

        Args:
            input_lines: An iterable of lines
        """
        self._buffer = None
        for line in input_lines:
            self._process_line(line)

    def process_and_buffer(self, input_lines):
        """ Reads an iterable of lines and writes to buffer.

        Pinball magic properties will be still written to the output rather
        than to the buffer.

        Args:
            input_lines: An iterable of lines
        """
        self._buffer = ''
        for line in input_lines:
            self._process_line(line)
        return self._buffer



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Logic handling reading logs from subprocess pipe"""

import os
from StringIO import StringIO


__author__ = 'Jooseong Kim'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class BufferedLineReader(object):
    """Reads text lines from a file object. The input stream may be buffered to
       form a line.

    This class encapsulates a text input stream that returns text lines.
    Python's file object readline() makes a blocking read() system call until
    it sees a new line delimiter in order to return a properly terminated text
    line. When readline() reads from a file descriptor on a file with limited
    buffer, the blocking system call may cause unexpected behavior. This class
    uses non-blocking read (os.read) calls so that it returns once it reads as
    many bytes that can be read from the buffer.

    This class was introduced to prevent the parent and child process
    communicating on pipes from deadlocking. The specific sequence of events
    that cause the deadlock is as follows:
    1. The child process writes a text string without the new line delimiter
       into the stdout pipe and flushes the buffer. As a result, the other end
       of the pipe becomes readable.
    2. select() returns and the pipe for the child's stdout becomes readable.
       The parent tries to read until a new line delimiter is seen, so it will
       block until the child process writes a new line delimiter intothe pipe.
    3. The child process writes a text string longer than the pipe buffer into
       stderr and flushes.

    As a result, the parent process is blocked on a read() on the stdout pipe
    from the child and waits for the child to write a new line delimiter into
    stdout. The child process is blocked on a write() on stderr and waits for
    the parent to read from the stderr pipe.

    Since the parent and child is waiting for each other on an action, this
    cyclic wait cannot be broken.
    """

    # default amount of bytes to read from the file object
    _DEFAULT_READ_SIZE = 1 << 11
    # default maximum size of the buffer we hold
    _DEFAULT_BUFFER_SIZE = 1 << 14

    def __init__(self, input_file, read_size=_DEFAULT_READ_SIZE,
                 max_buffer_size=_DEFAULT_BUFFER_SIZE):
        self._input_file = input_file
        self._read_size = read_size
        self._max_buffer_size = max_buffer_size

        self._reset_buffer()
        self._file_eof = False

    def _reset_buffer(self, initial_content=''):
        """Resets the buffer with initial_content
        """
        self._strio_buffer = StringIO()
        self._buffer_size = 0

        if initial_content:
            self._strio_buffer.write(initial_content)
            self._buffer_size = len(initial_content)

    def eof(self):
        """Returns whether EOF of the file is reached.

        Returns:
            True, if EOF of the file is reached.
            False, otherwise.
        """
        return self._file_eof

    def readlines(self):
        """Reads lines from a file.

        Reads new line (\n) terminated lines and maintains a buffer of text
        that is not yet terminated by a new line.

        Returns:
            list of one incomplete line, if the buffer is full.
            list of full lines read from the file, if such line exists.
            [], if no line could be formed.
        """

        # return empty list if EOF was already reached
        if self.eof():
            return []

        read_buffer = os.read(self._input_file.fileno(),
                              min(self._read_size,
                                  self._max_buffer_size - self._buffer_size))

        # empty string means we reached EOF
        if not read_buffer:
            self._file_eof = True

            # the buffer may have bytes not terminated by a new line, so return
            # the whole buffer.
            str_buffer = self._strio_buffer.getvalue()

            if str_buffer:
                return [str_buffer]
            else:
                return []

        self._strio_buffer.write(read_buffer)
        self._buffer_size += len(read_buffer)

        # new line is not found. There are two possible outcomes:
        # 1. Return empty list, if we still have space in the buffer. The text
        #    already read-in will be kept in _strio_buffer.
        # 2. Return a list with the content of the buffer if we reached the
        #    buffer size limit.
        if read_buffer.find('\n') < 0:
            if self._buffer_size < self._max_buffer_size:
                return []
            else:
                incomplete_line = self._strio_buffer.getvalue()
                self._reset_buffer()
                return [incomplete_line]

        str_buffer = self._strio_buffer.getvalue()

        # there is at least one full line in the buffer
        lines = []
        offset = 0
        while offset < len(str_buffer):
            newline_index = str_buffer.find('\n', offset)

            # no more new lines, so break out
            if newline_index < 0:
                break

            lines.append(str_buffer[offset:newline_index + 1])

            offset = newline_index + 1

        self._reset_buffer(str_buffer[offset:])

        return lines



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command line tool to interact with token master.

TODO(pawel): extend commands in this file to interact not only with the master
but also with the store.

TODO(mao): Make sure proper django setting when interact with store.
"""
import argparse
import sys

from pinball.config.pinball_config import PinballConfig
from pinball.config.utils import token_to_str
from pinball.master.factory import Factory
from pinball.master.thrift_lib.ttypes import GroupRequest
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import Token
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.tools.base import Command
from pinball.tools.base import CommandException
from pinball.tools.base import confirm
from pinball.workflow.name import Name


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


def _get_tokens(prefix, recursive, client):
    """Get tokens for a given name prefix.

    Args:
        prefix: The token name prefix to match.
        recursive: If False, only token with name fully matching the prefix
            will be retrieved.  Otherwise, all tokens with names starting with
            the prefix will be retrieved.
        client: The client to use when communicating with the master.
    Returns:
        List of tokens matching a given prefix.
    """
    result = []
    query = Query(namePrefix=prefix)
    request = QueryRequest(queries=[query])
    response = client.query(request)
    if response.tokens:
        assert len(response.tokens) == 1
        for token in response.tokens[0]:
            if recursive or token.name == prefix:
                result.append(token)
                if not recursive:
                    return result
    return result


class Cat(Command):
    """Show token content."""
    def __init__(self):
        self._prefix = None
        self._recursive = None

    def prepare(self, options):
        self._recursive = options.recursive
        if len(options.command_args) != 1:
            raise CommandException('cat command takes a token name prefix '
                                   'argument')
        self._prefix = options.command_args[0]

    def execute(self, client, store):
        output = ''
        tokens = _get_tokens(self._prefix, self._recursive, client)
        if not tokens:
            output += 'total 0\n'
        else:
            output += 'total %d\n' % len(tokens)
            for token in tokens:
                output += '%s\n' % token_to_str(token)
        return output


class Ls(Command):
    """List tokens in the master."""
    def __init__(self):
        self._prefix = None
        self._recursive = None

    def prepare(self, options):
        self._recursive = options.recursive
        if len(options.command_args) != 1:
            raise CommandException('ls command takes a token name prefix '
                                   'argument')
        self._prefix = options.command_args[0]

    def execute(self, client, store):
        output = ''
        if self._recursive:
            suffix = None
        else:
            suffix = Name.DELIMITER
        request = GroupRequest(namePrefix=self._prefix, groupSuffix=suffix)
        response = client.group(request)
        if not response.counts:
            output += 'total 0\n'
        else:
            output += 'total %d\n' % len(response.counts)
            for group in sorted(response.counts.keys()):
                output += '%s [%d token(s)]\n' % (group,
                                                  response.counts[group])
        return output


class Rm(Command):
    """Remove tokens from the master."""
    def __init__(self):
        self._force = None
        self._prefix = None
        self._recursive = None

    def prepare(self, options):
        self._recursive = options.recursive
        if len(options.command_args) != 1:
            raise CommandException('rm command takes a name prefix argument')
        self._prefix = options.command_args[0]
        self._force = options.force

    def execute(self, client, store):
        output = ''
        tokens = _get_tokens(self._prefix, self._recursive, client)
        deleted = 0
        if not tokens:
            output += 'no tokens found\n'
        else:
            print 'removing:'
            for token in tokens:
                print '\t%s' % token.name
            if self._force or confirm('remove %d tokens' % len(tokens)):
                request = ModifyRequest(deletes=tokens)
                client.modify(request)
                deleted = len(tokens)
        output += 'removed %d token(s)\n' % deleted
        return output


class Update(Command):
    """Insert or update tokens in the master."""
    def __init__(self):
        self._data = None
        self._owner = None
        self._expiration_time = None
        self._version = None
        self._name = None

    def prepare(self, options):
        if not options.name:
            raise CommandException('update command requires token name')
        self._name = options.name
        if options.version:
            self._version = options.version
        else:
            self._version = None
        if ((options.owner and not options.expiration_time) or
                (not options.owner and options.expiration_time)):
            raise CommandException('if either of owner and expiration_time is '
                                   'set, then the other must be set as well')
        if options.owner:
            self._owner = options.owner
        else:
            self._owner = None
        if options.expiration_time:
            self._expiration_time = options.expiration_time
        else:
            self._expiration_time = None
        self._priority = options.priority
        if options.data:
            self._data = options.data
        else:
            self._data = None
        if options.command_args:
            raise CommandException('update command does not take positional '
                                   'arguments')

    def execute(self, client, store):
        output = ''
        token = Token(name=self._name,
                      version=self._version,
                      owner=self._owner,
                      expirationTime=self._expiration_time,
                      priority=self._priority,
                      data=self._data)
        request = ModifyRequest(updates=[token])
        response = client.modify(request)
        assert len(response.updates) == 1
        if token.version is None:
            action = 'inserted'
        else:
            action = 'updated'
        output += '%s %s\n' % (action, str(response.updates[0]))
        output += 'updated 1 token\n'
        return output


_COMMANDS = {'cat': Cat, 'ls': Ls, 'rm': Rm, 'update': Update}


def main():
    parser = argparse.ArgumentParser(
        description='Interact with Pinball master server.')
    parser.add_argument('-p',
                        '--port',
                        dest='port',
                        type=int,
                        default=PinballConfig.MASTER_PORT,
                        help='port of the pinball master server')
    parser.add_argument('-s',
                        '--host',
                        dest='host',
                        default='localhost',
                        help='hostname of the pinball master server')
    parser.add_argument('-f',
                        '--force',
                        dest='force',
                        action='store_true',
                        default=False,
                        help='do not ask for confirmation')
    parser.add_argument('-r',
                        '--recursive',
                        dest='recursive',
                        action='store_true',
                        default=False,
                        help='perform the operation recursively')
    parser.add_argument('-n',
                        '--name',
                        dest='name',
                        help='token name')
    parser.add_argument('-v',
                        '--version',
                        dest='version',
                        type=int,
                        help='token version')
    parser.add_argument('-o',
                        '--owner',
                        dest='owner',
                        help='token owner; must be provided if '
                             'expiration_time is set')
    parser.add_argument('-t',
                        '--expiration_time',
                        dest='expiration_time',
                        type=int,
                        help='ownership expiration time in seconds since '
                             'epoch; must be provided if owner is set')
    parser.add_argument('-d',
                        '--data',
                        dest='data',
                        help='token data')
    parser.add_argument('-i',
                        '--priority',
                        dest='priority',
                        type=float,
                        default=0,
                        help='token priority')
    parser.add_argument('command',
                        choices=_COMMANDS.keys(),
                        help='command name')
    parser.add_argument('command_args',
                        nargs='*')
    options = parser.parse_args(sys.argv[1:])

    command = _COMMANDS[options.command]()
    command.prepare(options)
    factory = Factory(master_hostname=options.host, master_port=options.port)
    client = factory.get_client()
    print command.execute(client, None)


if __name__ == '__main__':
    main()



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for the token hierarchy inspector."""
import copy
import mock
import pickle
import time
import unittest

from pinball.master.factory import Factory
from pinball.workflow.event import Event
from pinball.workflow.name import Name
from pinball.workflow.job import ShellJob
from pinball.workflow.worker import Worker
from pinball.master.thrift_lib.ttypes import GroupRequest
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import Token
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.workflow.job_executor import ExecutionRecord
from tests.pinball.persistence.ephemeral_store import EphemeralStore
from pinball.workflow.signaller import Signal


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class WorkerTestCase(unittest.TestCase):
    def setUp(self):
        self._factory = Factory()
        self._store = EphemeralStore()
        self._factory.create_master(self._store)
        self._emailer = mock.Mock()
        self._worker = Worker(self._factory.get_client(), self._store,
                              self._emailer)
        self._client = self._factory.get_client()

    def _get_parent_job_token(self):
        name = Name(workflow='some_workflow',
                    instance='12345',
                    job_state=Name.WAITING_STATE,
                    job='parent_job')
        job = ShellJob(name=name.job,
                       inputs=[Name.WORKFLOW_START_INPUT],
                       outputs=['child_job'],
                       command='echo parent',
                       emails=['some_email@pinterest.com'])
        return Token(name=name.get_job_token_name(), data=pickle.dumps(job))

    def _get_child_job_token(self):
        name = Name(workflow='some_workflow',
                    instance='12345',
                    job_state=Name.WAITING_STATE,
                    job='child_job')
        job = ShellJob(name=name.job,
                       inputs=['parent_job'],
                       outputs=[],
                       command='echo child',
                       emails=['some_email@pinterest.com'])
        return Token(name=name.get_job_token_name(), data=pickle.dumps(job))

    def _post_job_tokens(self):
        """Add waiting job tokens to the master."""
        request = ModifyRequest(updates=[])
        request.updates.append(self._get_parent_job_token())
        request.updates.append(self._get_child_job_token())
        self._client.modify(request)

    def _post_workflow_start_event_token(self):
        name = Name(workflow='some_workflow',
                    instance='12345',
                    job='parent_job',
                    input_name=Name.WORKFLOW_START_INPUT,
                    event='workflow_start_event')
        event = Event(creator='SimpleWorkflowTest')
        token = Token(name=name.get_event_token_name(),
                      data=pickle.dumps(event))
        request = ModifyRequest(updates=[token])
        self._client.modify(request)

    def _verify_token_names(self, names):
        request = GroupRequest(namePrefix='/workflow/')
        response = self._client.group(request)
        names = sorted(names)
        counts = sorted(response.counts.keys())
        self.assertEqual(names, counts)

    def _verify_archived_token_names(self, names):
        active_tokens = self._store.read_active_tokens()
        all_tokens = self._store.read_tokens()
        archived_token_names = []
        for token in all_tokens:
            if not token in active_tokens:
                archived_token_names.append(token.name)
        names = sorted(names)
        archived_token_names = sorted(archived_token_names)
        self.assertEqual(names, archived_token_names)

    def _get_token(self, name):
        query = Query(namePrefix=name)
        request = QueryRequest(queries=[query])
        response = self._client.query(request)
        self.assertEqual(1, len(response.tokens))
        self.assertEqual(1, len(response.tokens[0]))
        return response.tokens[0][0]

    def _get_stored_token(self, name):
        tokens = self._store.read_tokens(name_prefix=name)
        self.assertEqual(1, len(tokens))
        return tokens[0]

    def _verify_parent_job_waiting(self):
        token_names = [
            Name(workflow='some_workflow',
                 instance='12345',
                 job_state=Name.WAITING_STATE,
                 job='parent_job').get_job_token_name(),
            Name(workflow='some_workflow',
                 instance='12345',
                 job_state=Name.WAITING_STATE,
                 job='child_job').get_job_token_name(),
            Name(workflow='some_workflow',
                 instance='12345',
                 job='parent_job',
                 input_name=Name.WORKFLOW_START_INPUT,
                 event='workflow_start_event').get_event_token_name()]
        self._verify_token_names(token_names)

    def _verify_parent_job_runnable(self):
        token_names = [Name(workflow='some_workflow',
                            instance='12345',
                            job_state=Name.RUNNABLE_STATE,
                            job='parent_job').get_job_token_name(),
                       Name(workflow='some_workflow',
                            instance='12345',
                            job_state=Name.WAITING_STATE,
                            job='child_job').get_job_token_name()]
        self._verify_token_names(token_names)

    def test_get_triggering_events(self):
        self.assertEqual([], Worker._get_triggering_events([]))

        self.assertEqual(['a'], Worker._get_triggering_events([['a']]))

        events = Worker._get_triggering_events([['a', 'b']])
        self.assertTrue(events == ['a'] or events == ['b'])

        events = Worker._get_triggering_events([['a', 'b'], ['1', '2']])
        self.assertTrue(events == ['a', '1'] or events == ['a', '2'] or
                        events == ['b', '1'] or events == ['b', '2'])

    def test_move_job_token_to_runnable(self):
        self._post_job_tokens()
        self._post_workflow_start_event_token()
        job_name = Name(workflow='some_workflow',
                        instance='12345',
                        job_state=Name.WAITING_STATE,
                        job='parent_job')
        job_token = self._get_token(job_name.get_job_token_name())
        event_name = Name(workflow='some_workflow',
                          instance='12345',
                          job='parent_job',
                          input_name=Name.WORKFLOW_START_INPUT,
                          event='workflow_start_event')
        event_token = self._get_token(event_name.get_event_token_name())
        self._worker._move_job_token_to_runnable(job_token, [event_token])
        # Event token should have been removed and the parent job should be
        # runnable.
        self._verify_parent_job_runnable()

    def test_make_job_runnable(self):
        self._post_job_tokens()
        self._post_workflow_start_event_token()

        parent_job_name = Name(workflow='some_workflow',
                               instance='12345',
                               job_state=Name.WAITING_STATE,
                               job='parent_job').get_job_token_name()
        child_job_name = Name(workflow='some_workflow',
                              instance='12345',
                              job_state=Name.WAITING_STATE,
                              job='child_job').get_job_token_name()

        parent_job_token = self._get_token(parent_job_name)
        child_job_token = self._get_token(child_job_name)

        self._worker._make_job_runnable(child_job_token)
        # Child job is missing triggering tokens so it cannot be made runnable.
        self._verify_parent_job_waiting()

        self._worker._make_job_runnable(parent_job_token)
        # Parent job has all triggering tokens so it can be made runnable.
        self._verify_parent_job_runnable()

    def test_make_runnable(self):
        self._post_job_tokens()
        self._post_workflow_start_event_token()

        self._worker._make_runnable('some_other_workflow', '12345')
        # Workflow instance does not exist so nothing should have changed.
        self._verify_parent_job_waiting()

        self._worker._make_runnable('some_workflow', 'some_other_instance')
        # Workflow instance does not exist so nothing should have changed.
        self._verify_parent_job_waiting()

        self._worker._make_runnable('some_workflow', '12345')
        self._verify_parent_job_runnable()

    def test_own_runnable_job_token(self):
        self._post_job_tokens()

        self._worker._own_runnable_job_token()
        # Event token is not present so nothing should have changed.
        token_names = [Name(workflow='some_workflow',
                            instance='12345',
                            job_state=Name.WAITING_STATE,
                            job='parent_job').get_job_token_name(),
                       Name(workflow='some_workflow',
                            instance='12345',
                            job_state=Name.WAITING_STATE,
                            job='child_job').get_job_token_name()]
        self._verify_token_names(token_names)
        self.assertIsNone(self._worker._owned_job_token)

        self._post_workflow_start_event_token()
        self._worker._own_runnable_job_token()
        # Worker should now own a runnable job token.
        self._verify_parent_job_runnable()
        parent_token = self._get_token(
            Name(workflow='some_workflow',
                 instance='12345',
                 job_state=Name.RUNNABLE_STATE,
                 job='parent_job').get_job_token_name())
        self.assertEqual(parent_token, self._worker._owned_job_token)

    def _add_history_to_owned_token(self):
        job = pickle.loads(self._worker._owned_job_token.data)
        execution_record = ExecutionRecord(start_time=123456,
                                           end_time=1234567,
                                           exit_code=0)
        job.history.append(execution_record)
        self._worker._owned_job_token.data = pickle.dumps(job)

    def test_get_output_event_tokens(self):
        self._post_job_tokens()
        self._post_workflow_start_event_token()
        self._worker._own_runnable_job_token()
        self.assertIsNotNone(self._worker._owned_job_token)

        job = pickle.loads(self._worker._owned_job_token.data)
        execution_record = ExecutionRecord(start_time=123456,
                                           end_time=1234567,
                                           exit_code=0)
        job.history.append(execution_record)

        event_tokens = self._worker._get_output_event_tokens(job)
        self.assertEqual(1, len(event_tokens))
        event_token_name = Name.from_event_token_name(event_tokens[0].name)
        expected_prefix = Name(workflow='some_workflow',
                               instance='12345',
                               job='child_job',
                               input_name='parent_job').get_input_prefix()
        self.assertEqual(expected_prefix, event_token_name.get_input_prefix())

    def test_move_job_token_to_waiting(self):
        self._post_job_tokens()
        self._post_workflow_start_event_token()
        self._worker._own_runnable_job_token()

        job = pickle.loads(self._worker._owned_job_token.data)
        execution_record = ExecutionRecord(start_time=123456,
                                           end_time=1234567,
                                           exit_code=0)
        job.history.append(execution_record)
        self._worker._owned_job_token.data = pickle.dumps(job)

        self._worker._move_job_token_to_waiting(job, True)

        parent_token = self._get_token(
            Name(workflow='some_workflow',
                 instance='12345',
                 job_state=Name.WAITING_STATE,
                 job='parent_job').get_job_token_name())
        job = pickle.loads(parent_token.data)
        self.assertEqual(1, len(job.history))
        self.assertEqual(execution_record.start_time,
                         job.history[0].start_time)

    def test_keep_job_token_in_runnable(self):
        self._post_job_tokens()
        self._post_workflow_start_event_token()
        self._worker._own_runnable_job_token()

        job = pickle.loads(self._worker._owned_job_token.data)
        job.history.append('some_historic_record')

        self._worker._keep_job_token_in_runnable(job)

        self._verify_parent_job_runnable()
        parent_token = self._get_token(
            Name(workflow='some_workflow',
                 instance='12345',
                 job_state=Name.RUNNABLE_STATE,
                 job='parent_job').get_job_token_name())
        job = pickle.loads(parent_token.data)
        self.assertEqual(1, len(job.history))
        self.assertEqual('some_historic_record', job.history[0])

    @staticmethod
    def _from_job(workflow, instance, job_name, job, data_builder, emailer):
        execution_record = ExecutionRecord(start_time=123456,
                                           end_time=1234567,
                                           exit_code=0)
        executed_job = copy.copy(job)
        executed_job.history.append(execution_record)
        job_executor = mock.Mock()
        job_executor.job = executed_job
        job_executor.prepare.return_value = True
        job_executor.execute.return_value = True
        return job_executor

    @mock.patch('pinball.workflow.worker.JobExecutor')
    def test_execute_job(self, job_executor_mock):
        self._post_job_tokens()
        self._post_workflow_start_event_token()
        self._worker._own_runnable_job_token()
        job_executor = mock.Mock()
        job_executor_mock.from_job.return_value = job_executor

        job_executor_mock.from_job.side_effect = WorkerTestCase._from_job

        self._worker._execute_job()

        self.assertIsNone(self._worker._owned_job_token)
        parent_token = self._get_token(
            Name(workflow='some_workflow',
                 instance='12345',
                 job_state=Name.WAITING_STATE,
                 job='parent_job').get_job_token_name())
        job = pickle.loads(parent_token.data)
        self.assertEqual(1, len(job.history))
        execution_record = job.history[0]
        self.assertEqual(0, execution_record.exit_code)
        self.assertEqual(1234567, execution_record.end_time)

    def test_send_instance_end_email(self):
        data_builder = mock.Mock()
        self._worker._data_builder = data_builder

        schedule_data = mock.Mock()
        schedule_data.emails = ['some_email@pinterest.com']
        data_builder.get_schedule.return_value = schedule_data

        instance_data = mock.Mock()
        data_builder.get_instance.return_value = instance_data

        job_data = mock.Mock()
        data_builder.get_jobs.return_value = [job_data]

        self._worker._send_instance_end_email('some_workflow', '12345')

        self._emailer.send_instance_end_message.assert_called_once_with(
            ['some_email@pinterest.com'], instance_data, [job_data])

    def test_send_job_failure_emails(self):
        self._post_job_tokens()
        self._post_workflow_start_event_token()
        self._worker._own_runnable_job_token()

        job = pickle.loads(self._worker._owned_job_token.data)
        job.history.append('some_historic_record')
        executor = mock.Mock()
        self._worker._executor = executor
        executor.job = job

        data_builder = mock.Mock()
        self._worker._data_builder = data_builder

        schedule_data = mock.Mock()
        schedule_data.emails = ['some_other_email@pinterest.com']
        data_builder.get_schedule.return_value = schedule_data

        execution_data = mock.Mock()
        data_builder.get_execution.return_value = execution_data

        self._worker._send_job_failure_emails(True)

        self._emailer.send_job_execution_end_message.assert_any_call(
            ['some_email@pinterest.com',
             'some_other_email@pinterest.com'], execution_data)

    @mock.patch('pinball.workflow.worker.JobExecutor')
    def test_run(self, job_executor_mock):
        self._post_job_tokens()
        self._post_workflow_start_event_token()

        job_executor_mock.from_job.side_effect = WorkerTestCase._from_job

        self._worker._test_only_end_if_no_runnable = True
        self._worker.run()
        with mock.patch('pinball.workflow.archiver.time') as time_patch:
            # add one day
            time_patch.time.return_value = time.time() + 24 * 60 * 60
            self._worker.run()

        parent_job_token_name = Name(workflow='some_workflow',
                                     instance='12345',
                                     job_state=Name.WAITING_STATE,
                                     job='parent_job').get_job_token_name()
        child_job_token_name = Name(workflow='some_workflow',
                                    instance='12345',
                                    job_state=Name.WAITING_STATE,
                                    job='child_job').get_job_token_name()
        signal_string = Signal.action_to_string(Signal.ARCHIVE)
        signal_token_name = Name(workflow='some_workflow',
                                 instance='12345',
                                 signal=signal_string).get_signal_token_name()

        token_names = [parent_job_token_name,
                       child_job_token_name,
                       signal_token_name]
        self._verify_archived_token_names(token_names)

        self.assertEqual(2, job_executor_mock.from_job.call_count)

        parent_token = self._get_stored_token(parent_job_token_name)
        job = pickle.loads(parent_token.data)
        self.assertEqual(1, len(job.history))
        execution_record = job.history[0]
        self.assertEqual(0, execution_record.exit_code)
        self.assertEqual(1234567, execution_record.end_time)

        child_token = self._get_stored_token(child_job_token_name)
        job = pickle.loads(child_token.data)
        self.assertEqual(1, len(job.history))
        execution_record = job.history[0]
        self.assertEqual(0, execution_record.exit_code)
        self.assertEqual(1234567, execution_record.end_time)

        signal_token = self._get_stored_token(signal_token_name)
        signal = pickle.loads(signal_token.data)
        self.assertEqual(Signal.ARCHIVE, signal.action)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation tests for pinball_util tool."""
import collections
import mock
import unittest

from pinball.tools.pinball_util import Cat
from pinball.tools.pinball_util import Ls
from pinball.tools.pinball_util import Rm
from pinball.tools.pinball_util import Update
from pinball.master.thrift_lib.ttypes import GroupRequest
from pinball.master.thrift_lib.ttypes import GroupResponse
from pinball.master.thrift_lib.ttypes import ModifyRequest
from pinball.master.thrift_lib.ttypes import ModifyResponse
from pinball.master.thrift_lib.ttypes import Query
from pinball.master.thrift_lib.ttypes import QueryRequest
from pinball.master.thrift_lib.ttypes import QueryResponse
from pinball.master.thrift_lib.ttypes import Token


__author__ = 'Pawel Garbacki'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


class CatTestCase(unittest.TestCase):
    def test_empty(self):
        Options = collections.namedtuple('args', 'recursive command_args')
        options = Options(recursive=False, command_args=['/some_path'])
        command = Cat()
        command.prepare(options)
        client = mock.Mock()
        response = QueryResponse()
        client.query.return_value = response

        output = command.execute(client, None)
        query = Query(namePrefix='/some_path')
        request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(request)
        self.assertEqual('total 0\n', output)

    def test_recursive(self):
        Options = collections.namedtuple('args', 'recursive command_args')
        options = Options(recursive=True, command_args=['/some_path'])
        command = Cat()
        command.prepare(options)

        client = mock.Mock()
        token = Token(version=10,
                      name='/some_path/some_token',
                      owner='some_owner',
                      expirationTime=10,
                      data='some_data')
        query_response = QueryResponse(tokens=[[token]])
        client.query.return_value = query_response

        output = command.execute(client, None)

        query = Query(namePrefix='/some_path')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)

        self.assertEqual('total 1\nToken(version=10, owner=some_owner, '
                         'expirationTime=1970-01-01 00:00:10 UTC, '
                         'priority=0.000000, name=/some_path/some_token, '
                         'data=some_data)\n',
                         output)


class LsTestCase(unittest.TestCase):
    def test_empty(self):
        Options = collections.namedtuple('args', 'recursive command_args')
        options = Options(recursive=False, command_args=['/some_path'])
        command = Ls()
        command.prepare(options)
        client = mock.Mock()
        response = GroupResponse()
        client.group.return_value = response

        output = command.execute(client, None)

        request = GroupRequest(namePrefix='/some_path', groupSuffix='/')
        client.group.assert_called_once_with(request)
        self.assertEqual('total 0\n', output)

    def test_recursive(self):
        Options = collections.namedtuple('args', 'recursive command_args')
        options = Options(recursive=True, command_args='/')
        command = Ls()
        command.prepare(options)
        client = mock.Mock()
        # Respond 10, and that should come in the output of the executed
        # command.
        response = GroupResponse(counts={'/some_path': 10})
        client.group.return_value = response

        output = command.execute(client, None)

        self.assertEqual('total 1\n/some_path [10 token(s)]\n', output)


class RmTestCase(unittest.TestCase):
    def test_empty(self):
        Options = collections.namedtuple('args',
                                         'recursive force command_args')
        options = Options(recursive=False, force=True,
                          command_args=['/some_path'])
        command = Rm()
        command.prepare(options)
        client = mock.Mock()
        response = QueryResponse()
        client.query.return_value = response

        output = command.execute(client, None)

        query = Query(namePrefix='/some_path')
        request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(request)
        self.assertEqual('no tokens found\nremoved 0 token(s)\n', output)

    def test_recursive(self):
        Options = collections.namedtuple('args',
                                         'recursive force command_args')
        options = Options(recursive=True, force=True,
                          command_args=['/some_path'])
        command = Rm()
        command.prepare(options)

        client = mock.Mock()
        token = Token(version=10,
                      name='/some_path/some_token',
                      owner='some_owner',
                      expirationTime=10,
                      data='some_data')
        query_response = QueryResponse(tokens=[[token]])
        client.query.return_value = query_response

        modify_response = ModifyResponse()
        client.modify.return_value = modify_response

        output = command.execute(client, None)

        query = Query(namePrefix='/some_path')
        query_request = QueryRequest(queries=[query])
        client.query.assert_called_once_with(query_request)

        modify_request = ModifyRequest(deletes=[token])
        client.modify.assert_called_once_with(modify_request)

        self.assertEqual('removed 1 token(s)\n', output)


class UpdateTestCase(unittest.TestCase):
    def test_insert(self):
        Options = collections.namedtuple('args', 'name version owner '
                                         'expiration_time priority data '
                                         'command_args')
        options = Options(name='/some_path/some_token',
                          version=None,
                          owner=None,
                          expiration_time=None,
                          priority=0,
                          data=None,
                          command_args=None)
        command = Update()
        command.prepare(options)
        client = mock.Mock()
        output_token = Token(version=10,
                             name='/some_path/some_token',
                             owner='some_owner',
                             expirationTime=10,
                             data='some_data')
        response = ModifyResponse(updates=[output_token])
        client.modify.return_value = response

        output = command.execute(client, None)

        input_token = Token(name='/some_path/some_token')
        request = ModifyRequest(updates=[input_token])
        client.modify.assert_called_once_with(request)
        self.assertEqual('inserted %s\nupdated 1 token\n' % str(output_token),
                         output)

    def test_update(self):
        Options = collections.namedtuple('args', 'name version owner '
                                         'expiration_time priority data '
                                         'command_args')
        options = Options(name='/some_path/some_token',
                          version=10,
                          owner='some_other_owner',
                          expiration_time=100,
                          priority=10,
                          data='some_other_data',
                          command_args=None)
        command = Update()
        command.prepare(options)
        client = mock.Mock()
        output_token = Token(version=11,
                             name='/some_path/some_token',
                             owner='some_other_owner',
                             expirationTime=100,
                             priority=10,
                             data='some_other_data')
        response = ModifyResponse(updates=[output_token])
        client.modify.return_value = response

        output = command.execute(client, None)

        input_token = Token(version=10,
                            name='/some_path/some_token',
                            owner='some_other_owner',
                            expirationTime=100,
                            priority=10,
                            data='some_other_data')
        request = ModifyRequest(updates=[input_token])
        client.modify.assert_called_once_with(request)
        self.assertEqual('updated %s\nupdated 1 token\n' % str(output_token),
                         output)



# Copyright 2015, Pinterest, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for interacting with Amazon s3 service.

Please call config_s3_utils() first before calling any other methods that will
interact with Amazon s3 service in this module.

For more information about, please check:
    - s3: http://aws.amazon.com/s3/
    - boto: http://aws.amazon.com/sdk-for-python/

TODO(csliu): change the utils from c-style to OO style to reuse connection etc.
"""

import os
import re

import boto

from pinball_ext.common.utils import get_logger


__author__ = 'Mao Ye, Changshu Liu'
__copyright__ = 'Copyright 2015, Pinterest, Inc.'
__credits__ = [__author__]
__license__ = 'Apache'
__version__ = '2.0'


AWS_ACCESS_KEY_ID = ''
AWS_SECRET_ACCESS_KEY = ''
if not boto.config.has_section('Boto'):
    boto.config.add_section('Boto')
    boto.config.set('Boto', 'http_socket_timeout', '180')

LOG = get_logger('pinball_ext.common.s3_utils')


def config_s3_utils(aws_access_key_id, aws_secret_access_key):
    """Config core parameters for s3_utils module.

    Args:
        aws_access_key_id: first parameter passed to boto.connect_s3()
        aws_secrete_access_key: second parameter passed to boto.connect_s3()

    Returns:
        None
    """
    assert aws_access_key_id
    assert aws_secret_access_key

    global AWS_ACCESS_KEY_ID
    global AWS_SECRET_ACCESS_KEY
    AWS_ACCESS_KEY_ID = aws_access_key_id
    AWS_SECRET_ACCESS_KEY = aws_secret_access_key


def parse_s3_path(s3_path):
    """Parse s3 directory into bucket name and remaining path.

    Args:
        s3_path: a s3 path following the s3n scheme.

    Returns:
        bucket_name, path_name
    """
    directory_regex = r's3n?://(?P<bucket_name>[\w-]+)/(?P<path_name>.+)'
    m = re.search(directory_regex, s3_path)
    if m:
        bucket_name = m.group('bucket_name')
        path_name = m.group('path_name')
        if bucket_name and path_name:
            return bucket_name, path_name
    return None, None


def get_s3_bucket(bucket_name):
    """Get the Boto s3 bucket reference for the given bucket_name.

    Args:
        bucket_name: name of s3 bucket.

    Returns
        s3 Boto bucket object.
    """
    assert AWS_ACCESS_KEY_ID
    assert AWS_SECRET_ACCESS_KEY
    connection = boto.connect_s3(AWS_ACCESS_KEY_ID,
                                 AWS_SECRET_ACCESS_KEY)
    assert connection
    bucket = connection.get_bucket(bucket_name, validate=False)
    return bucket


def get_s3_key(s3_path):
    """Get the Boto s3 key for given s3_path.

    Args:
        s3_path: a s3 path following the s3n scheme.

    Returns
        s3 Boto key object.
    """
    bucket_name, path = parse_s3_path(s3_path)
    bucket = get_s3_bucket(bucket_name)
    key = bucket.get_key(path)
    return key


def get_or_create_s3_key(s3_path):
    """Gets or creates a Boto reference to a s3 key.

    Usage:
        with get_or_create_key(s3_path) as key:
            key.set_contents_from_string(data)

    Args:
        s3_path - of the form 's3n://<bucket_name>/<path>

    Returns:
        the Boto key.
    """
    bucket_name, path = parse_s3_path(s3_path)
    bucket = get_s3_bucket(bucket_name)
    key = bucket.get_key(path)
    if not key:
        key = bucket.new_key(path)
    return key


def list_s3_directory(s3_directory):
    """List all keys under given s3_directory.

    Args:
        s3_directory: a s3 directory path following s3n scheme.

    Returns:
        list of names of s3 keys.
    """
    bucket_name, path_name = parse_s3_path(s3_directory)
    path_name = path_name if path_name[-1] == '/' else path_name + '/'
    bucket = get_s3_bucket(bucket_name)
    rs = bucket.list(prefix=path_name)
    s3_names = [key.name for key in rs]
    return s3_names


def s3_put_string(s3_location, data):
    """Uploads a data string to a particular location on S3.

    Args:
        s3_location - of the form 's3n://<bucket_name>/<path>
        data - a string
    """
    key = get_or_create_s3_key(s3_location)
    key.set_contents_from_string(data)


def delete_s3_directory(s3_directory):
    """Delete the given s3 directory."""
    bucket_name, rest_of_dir_path = parse_s3_path(s3_directory)
    bucket = get_s3_bucket(bucket_name)
    rest_of_dir_path = rest_of_dir_path if rest_of_dir_path[-1] == '/' \
        else rest_of_dir_path + '/'
    bucket.delete_keys(bucket.list(prefix=rest_of_dir_path))


def extract_file_name_from_s3_path(s3_path):
    """Extract file name from a full s3 path."""
    # is os.path.split(str(bucket_entry.name))[1] safe?
    # >>> os.path.split('/foo')
    # ('/', 'foo')
    # >>> os.path.split('/')
    # ('/', '')
    # >>> os.path.split('/foo/')
    # ('/foo', '')
    # So for '/' and '/foo/' the array element #2 will be empty.
    # We can assume S3 will not give us such entries since:
    # 1. the bucket is not empty so '/' will not exist;
    # 2. '/foo/' will not appear as there is no 'directory' on S3
    # and everything is a file.
    return os.path.split(str(s3_path))[1]
